#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.IO;
using Ami.Objects.Core.ReflectionModel;
using Ami.Objects.Relations;

namespace Ami.Objects.Attributes
{
    /// <summary>
    /// Marks a property as the domain property.
    /// </summary>
    [AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = false)]
    public class DomainAttribute : RelationKindAttribute
    {
        /// <summary>
        /// Implementation of <see cref="IDomainRelation{TDomain,TOwner}"/>.
        /// </summary>
        /// <typeparam name="TDomain">Domain type.</typeparam>
        /// <typeparam name="TRelationOwner">Relation owner type.</typeparam>
        public sealed class DomainRelation<TDomain, TRelationOwner> : ARelation<TRelationOwner>, IDomainRelation<TDomain, TRelationOwner>
            where TRelationOwner : class, IObject
            where TDomain : class, IDomain
        {
            /// <summary>
            /// Initializes a new instance of <see cref="DomainRelation{TDomain,TRelationOwner}"/>.
            /// </summary>
            /// <param name="name">Relation name.</param>
            /// <param name="domain">Domain of the <see cref="TRelationOwner"/>.</param>
            /// <param name="relationOwner">Relation owner.</param>
            public DomainRelation(string name, TDomain domain, TRelationOwner relationOwner)
                : base(name, relationOwner)
            {
                Domain = domain;
            }

            /// <summary>
            /// <see cref="IObject.Domain"/> typed value.
            /// </summary>
            public TDomain Domain { get; private set; }

            /// <summary>
            /// <see cref="IObject.Domain"/> value.
            /// </summary>
            IDomain IDomainRelation.Domain
            {
                get { return Domain; }
            }

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
            /// </summary>
            /// <filterpriority>2</filterpriority>
            public override void Dispose()
            {
                base.Dispose();
                Domain = null;
            }
        }

        /// <summary>
        /// Special <see cref="IDomainRelation"/> implementation for domain objects.
        /// </summary>
        public sealed class DomainDomainRelation<TDomain> : ARelation<TDomain>, IDomainRelation<TDomain, TDomain> where TDomain : class, IDomain
        {
            public DomainDomainRelation(string name, TDomain relationOwner)
                : base(name, relationOwner)
            {
            }

            /// <summary>
            /// <see cref="IObject.Domain"/> value.
            /// </summary>
            IDomain IDomainRelation.Domain
            {
                get { return RelationOwner; }
            }

            /// <summary>
            /// <see cref="IObject.Domain"/> typed value.
            /// </summary>
            /// 
            public TDomain Domain
            {
                get { return RelationOwner; }
            }
        }

        /// <summary>
        /// <see cref="DomainAttribute"/> kind.
        /// This can also implement Owner properties when they are also domain.
        /// </summary>
        public interface IKind : OwnerAttribute.IKind
        {
        }

        /// <summary>
        /// Order used by <see cref="AmiRelation.Comparer"/>.
        /// </summary>
        public override int Order
        {
            get { return int.MinValue; }
        }

        /// <summary>
        /// Get the kind of this attribute.
        /// This must return an interface.
        /// </summary>
        protected override Type GetKind
        {
            get { return typeof (IKind); }
        }

        /// <summary>
        /// Implements an <see cref="DomainAttribute"/>.
        /// </summary>
        /// <param name="relation">Relation to implement.</param>
        public override string Implement(AmiRelation relation)
        {
            return ImplementProperties(relation);
        }

        /// <summary>
        /// Create the corresponding relation instanciation.
        /// The text writer is like:
        /// <code>
        /// namespace N
        /// {
        ///     public class C
        ///     {
        ///         void CreateRelations()
        ///         {
        /// </code>
        /// </summary>
        /// <param name="relation">Relation to create.</param>
        /// <param name="domain">True if the relation to create is the object's domain relation.</param>
        public override string CreateRelation(AmiRelation relation, bool domain)
        {
            using (var code = new StringWriter())
            {
                if (relation.Type.IsDomainType)
                {
                    code.Write("new Ami.Objects.Attributes.DomainAttribute.DomainDomainRelation<");
                    code.Write(relation.Type.FullTypeName);
                    code.Write(">(");
                    code.Quote(relation.Name);
                    code.Write(", this)");
                }
                else if (!domain && relation.Type.OwnerRelation == relation)
                {
                    code.Write("new Ami.Objects.Attributes.OwnerAttribute.DomainAndOwnerRelation<");
                    code.WriteCSharpQualifiedName(relation.RelationType);
                    code.Write(", ");
                    code.Write(relation.Type.FullTypeName);
                    code.Write(">(");
                    code.Quote(() =>
                                   {
                                       code.Write(relation.Name);
                                       code.Write("Owner");
                                   });
                    code.Write(", this)");
                }
                else
                {
                    code.Write("new Ami.Objects.Attributes.DomainAttribute.DomainRelation<");
                    code.WriteCSharpQualifiedName(relation.RelationType);
                    code.Write(", ");
                    code.Write(relation.Type.FullTypeName);
                    code.Write(">(");
                    code.Quote(relation.Name);
                    code.Write(", domain, this)");
                }
                return code.ToString();
            }
        }

        /// <summary>
        /// Validates a <paramref name="relation"/> with this Kind.
        /// </summary>
        /// <param name="relation"><see cref="AmiRelation"/> to validate.</param>
        public override void Validate(AmiRelation relation)
        {
            base.Validate(relation);
            ValidateReadonlyProperties(relation);
            ValidateDomainType(relation);
        }

        private static void ValidateDomainType(AmiRelation relation)
        {
            if (relation.Type.IsDomainType && relation.RelationType != null && relation.RelationType.Inherits<IDomain>())
            {
                if (!relation.Type.ReflectedType.Inherits(relation.RelationType)) relation.Add(InfoId.DomainType, "Relation domain type {0} is incompatible with domain type", relation.RelationType.CSharpQualifiedName());
            }
        }

        /// <summary>
        /// Validates a <paramref name="type"/> with some relation of this Kind.
        /// </summary>
        /// <param name="type"><see cref="AmiType"/> to validate.</param>
        public override void Validate(AmiType type)
        {
            base.Validate(type);
            ValidateSingleRelationPerObject(type);
        }
    }
}