#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.IO;
using Ami.Objects.Core.ReflectionModel;
using Ami.Objects.Relations;

namespace Ami.Objects.Attributes
{
    /// <summary>
    /// Marks a property as the owner property.
    /// </summary>
    [AttributeUsage(AttributeTargets.Property, AllowMultiple = false, Inherited = false)]
    public class OwnerAttribute : RelationKindAttribute
    {
        /// <summary>
        /// Relation of <see cref="OwnerAttribute"/>.
        /// </summary>
        /// <typeparam name="TOwner">Type of <see cref="IObject"/> owner.</typeparam>
        /// <typeparam name="TRelationOwner">Type of <see cref="IObject"/> owned.</typeparam>
        public class OwnerRelation<TOwner, TRelationOwner> : ARelation<TRelationOwner>, IOwnerRelation<TOwner, TRelationOwner>
            where TRelationOwner : class, IObject
            where TOwner : IObject
        {
            /// <summary>
            /// Initializes a new instance of <see cref="OwnerRelation{TOwner,TRelationOwner}"/>.
            /// </summary>
            /// <param name="name">Relation name.</param>
            /// <param name="relationOwner">Relation owner.</param>
            public OwnerRelation(string name, TRelationOwner relationOwner)
                : base(name, relationOwner)
            {
            }

            /// <summary>
            /// <see cref="IObject"/> owned by <see cref="IRelation.RelationOwner"/>.
            /// This property is equivalent to <see cref="IRelation.RelationOwner"/>.
            /// </summary>
            IObject IOwnerRelation.Owned
            {
                get { return Owned; }
            }

            /// <summary>
            /// <see cref="IObject"/> owned by <see cref="IRelation.RelationOwner"/>.
            /// This property is equivalent to <see cref="IRelation{TOwner}.RelationOwner"/>.
            /// </summary>
            public TRelationOwner Owned
            {
                get { return RelationOwner; }
            }

            /// <summary>
            /// <see cref="ObjectOwner"/>.
            /// </summary>
            private TOwner objectOwner;

            /// <summary>
            /// Typed <see cref="IOwnerRelation.ObjectOwner"/>.
            /// </summary>
            public virtual TOwner ObjectOwner
            {
                get { return objectOwner; }
                set
                {
                    if (!ReferenceEquals(objectOwner, value))
                    {
                        var oldValue = objectOwner;
                        objectOwner = value;
                        {
                            var e = OnOwnerChangedTOwnerTOwnerRelation;
                            if (e != null) e(this, oldValue, objectOwner);
                        }
                        {
                            var e = OnOwnerChangedTOwner;
                            if (e != null) e(this, oldValue, objectOwner);
                        }
                        {
                            var e = OnOwnerChanged;
                            if (e != null) e(this, oldValue, objectOwner);
                        }
                    }
                }
            }

            /// <summary>
            /// <see cref="IObject.Owner"/> value.
            /// </summary>
            IObject IOwnerRelation.ObjectOwner
            {
                get { return ObjectOwner; }
                set { ObjectOwner = (TOwner) value; }
            }

            private event ObjectOwnerChanged<TOwner, TRelationOwner> OnOwnerChangedTOwnerTOwnerRelation;
            private event ObjectOwnerChanged<TOwner> OnOwnerChangedTOwner;
            private event ObjectOwnerChanged OnOwnerChanged;

            /// <summary>
            /// Occurs when <see cref="IOwnerRelation{TObjectOwner}.ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged<TOwner, TRelationOwner> IOwnerRelation<TOwner, TRelationOwner>.ObjectOwnerChanged
            {
                add { lock (OnOwnerChangedTOwnerTOwnerRelation) OnOwnerChangedTOwnerTOwnerRelation += value; }
                remove { lock (OnOwnerChangedTOwnerTOwnerRelation) OnOwnerChangedTOwnerTOwnerRelation -= value; }
            }

            /// <summary>
            /// Occurs when <see cref="IOwnerRelation{TObjectOwner}.ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged<TOwner> IOwnerRelation<TOwner>.ObjectOwnerChanged
            {
                add { lock (OnOwnerChangedTOwner) OnOwnerChangedTOwner += value; }
                remove { lock (OnOwnerChangedTOwner) OnOwnerChangedTOwner -= value; }
            }

            /// <summary>
            /// Occurs when <see cref="IOwnerRelation.ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged IOwnerRelation.ObjectOwnerChanged
            {
                add { lock (OnOwnerChanged) OnOwnerChanged += value; }
                remove { lock (OnOwnerChanged) OnOwnerChanged -= value; }
            }
        }

        /// <summary>
        /// Special <see cref="IOwnerRelation"/> implementation for domain objects.
        /// </summary>
        public sealed class DomainOwnerRelation<TObject> : ARelation<TObject>, IOwnerRelation<IObject, TObject> where TObject : class, IDomain
        {
            public DomainOwnerRelation(string name, TObject relationOwner)
                : base(name, relationOwner)
            {
            }

            /// <summary>
            /// <see cref="IObject"/> owned by <see cref="IRelation.RelationOwner"/>.
            /// This property is equivalent to <see cref="IRelation.RelationOwner"/>.
            /// </summary>
            IObject IOwnerRelation.Owned
            {
                get { return Owned; }
            }

            /// <summary>
            /// <see cref="IObject"/> owned by <see cref="IRelation.RelationOwner"/>.
            /// This property is equivalent to <see cref="IRelation{TRelationOwner}.RelationOwner"/>.
            /// </summary>
            public TObject Owned
            {
                get { return RelationOwner; }
            }

            /// <summary>
            /// Occurs when <see cref="IOwnerRelation{TObjectOwner}.ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged<IObject, TObject> IOwnerRelation<IObject, TObject>.ObjectOwnerChanged
            {
                add { }
                remove { }
            }

            /// <summary>
            /// Occurs when <see cref="ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged<IObject> IOwnerRelation<IObject>.ObjectOwnerChanged
            {
                add { }
                remove { }
            }

            /// <summary>
            /// Occurs when <see cref="ObjectOwner"/> has changed.
            /// </summary>
            event ObjectOwnerChanged IOwnerRelation.ObjectOwnerChanged
            {
                add { }
                remove { }
            }

            /// <summary>
            /// Typed <see cref="IOwnerRelation.ObjectOwner"/>.
            /// </summary>
            public IObject ObjectOwner
            {
                get { return null; }
                set { throw new InvalidOperationException("Domain cannot be owned"); }
            }

            /// <summary>
            /// <see cref="IObject.Owner"/> value.
            /// </summary>
            IObject IOwnerRelation.ObjectOwner
            {
                get { return ObjectOwner; }
                set { ObjectOwner = value; }
            }
        }

        /// <summary>
        /// Special <see cref="IOwnerRelation"/> implementation when object's onwer is also the domain.
        /// </summary>
        /// <typeparam name="TDomain"></typeparam>
        /// <typeparam name="TRelationOwner"></typeparam>
        public sealed class DomainAndOwnerRelation<TDomain, TRelationOwner> : OwnerRelation<TDomain, TRelationOwner>
            where TDomain : class, IDomain
            where TRelationOwner : class, IObject
        {
            /// <summary>
            /// Initializes a new instance of <see cref="DomainAndOwnerRelation{TDomain,TRelationOwner}"/>.
            /// </summary>
            /// <param name="name">Relation name.</param>
            /// <param name="relationOwner">Relation owner.</param>
            public DomainAndOwnerRelation(string name, TRelationOwner relationOwner)
                : base(name, relationOwner)
            {
            }

            /// <summary>
            /// Typed <see cref="IOwnerRelation.ObjectOwner"/>.
            /// </summary>
            public override TDomain ObjectOwner
            {
                get { return base.ObjectOwner; }
                set
                {
                    if (value != null && value != RelationOwner.Domain) throw new InvalidOperationException("Owner must be the object's domain");
                    base.ObjectOwner = value;
                }
            }
        }

        /// <summary>
        /// <see cref="OwnerAttribute"/> kind.
        /// </summary>
        public interface IKind
        {
        }

        /// <summary>
        /// Order used by <see cref="AmiRelation.Comparer"/>.
        /// </summary>
        public override int Order
        {
            get { return int.MinValue + 1; }
        }

        /// <summary>
        /// Get the kind of this attribute.
        /// This must return an interface.
        /// </summary>
        protected override Type GetKind
        {
            get { return typeof (IKind); }
        }

        /// <summary>
        /// Implements an <see cref="OwnerAttribute"/>.
        /// </summary>
        /// <param name="relation">Relation to implement.</param>
        public override string Implement(AmiRelation relation)
        {
            return "";
        }

        /// <summary>
        /// Create the corresponding relation instanciation.
        /// The text writer is like:
        /// <code>
        /// namespace N
        /// {
        ///     public class C
        ///     {
        ///         void CreateRelations()
        ///         {
        /// </code>
        /// </summary>
        /// <param name="relation">Relation to create.</param>
        /// <param name="domain">True if the relation to create is the object's domain relation.</param>
        public override string CreateRelation(AmiRelation relation, bool domain)
        {
            using (var code = new StringWriter())
            {
                if (relation.Type.IsDomainType)
                {
                    code.Write("new Ami.Objects.Attributes.OwnerAttribute.DomainOwnerRelation<");
                    code.Write(relation.Type.FullTypeName);
                    code.Write(">(");
                    code.Quote(relation.Name);
                    code.Write(", this)");
                }
                else
                {
                    code.Write("new Ami.Objects.Attributes.OwnerAttribute.OwnerRelation<");
                    code.WriteCSharpQualifiedName(relation.RelationType);
                    code.Write(", ");
                    code.Write(relation.Type.FullTypeName);
                    code.Write(">(");
                    code.Quote(relation.Name);
                    code.Write(", this)");
                }
                return code.ToString();
            }
        }

        /// <summary>
        /// Validates a <paramref name="relation"/> with this Kind.
        /// </summary>
        /// <param name="relation"><see cref="AmiRelation"/> to validate.</param>
        public override void Validate(AmiRelation relation)
        {
            base.Validate(relation);
            foreach (var property in relation.ReflectedProperties)
            {
                if (property.CanWrite) relation.Add(InfoId.ReadonlyRelation, "Property \"{0}\" should not have a setter", property.CSharpSignature());
            }
        }
    }
}