#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.IO;
using System.Linq;
using Ami.Objects.Core;
using Ami.Objects.Core.ReflectionModel;
using Ami.Objects.Relations;

namespace Ami.Objects.Attributes
{
    /// <summary>
    /// A kind of attribute is responsible for implementing a kind of property.
    /// The selected kind is the one with the most specialized <see cref="Type"/>.
    /// </summary>
    public abstract class RelationKindAttribute : Attribute
    {
        /// <summary>
        /// Abstract <see cref="IRelation"/> implementation.
        /// </summary>
        public abstract class ARelation : IRelation
        {
            /// <summary>
            /// Initializes a new instance of <see cref="ARelation"/> with the given <paramref name="name"/>.
            /// </summary>
            /// <param name="name">Relation name.</param>
            public ARelation(string name)
            {
                Name = name;
            }

            /// <summary>
            /// Initializes this relation.
            /// </summary>
            public virtual void Initialize()
            {
            }

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
            /// </summary>
            /// <filterpriority>2</filterpriority>
            public virtual void Dispose()
            {
                Name = null;
            }

            /// <summary>
            /// <see cref="IRelation"/>'s name.
            /// </summary>
            public string Name { get; private set; }

            /// <summary>
            /// <see cref="IObject"/> owning this relation.
            /// </summary>
            IObject IRelation.RelationOwner
            {
                get { return GetRelationOwner(); }
            }

            /// <summary>
            /// <see cref="IObject"/> owning this relation.
            /// </summary>
            protected abstract IObject GetRelationOwner();
        }

        /// <summary>
        /// Typed abstract implementation of <see cref="IRelation{T}"/>
        /// </summary>
        /// <typeparam name="TRelationOwner"></typeparam>
        public abstract class ARelation<TRelationOwner> : ARelation, IRelation<TRelationOwner> where TRelationOwner : class, IObject
        {
            /// <summary>
            /// Initializes a new instance of <see cref="ARelation{TRelationOwner}"/> with the given <paramref name="relationOwner"/>.
            /// </summary>
            /// <param name="name">Relation nanme.</param>
            /// <param name="relationOwner">Object owning this relation.</param>
            protected ARelation(string name, TRelationOwner relationOwner)
                : base(name)
            {
                RelationOwner = relationOwner;
            }

            /// <summary>
            /// <typeparam name="TRelationOwner"/> owning this relation.
            /// </summary>
            public TRelationOwner RelationOwner { get; private set; }

            /// <summary>
            /// <see cref="IObject"/> owning this relation.
            /// </summary>
            protected override IObject GetRelationOwner()
            {
                return RelationOwner;
            }

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
            /// </summary>
            /// <filterpriority>2</filterpriority>
            public override void Dispose()
            {
                base.Dispose();
                RelationOwner = null;
            }
        }

        /// <summary>
        /// Get the kind of this attribute.
        /// </summary>
        public Type Type
        {
            get
            {
                var type = GetKind;
                if (type == null) throw new InvalidOperationException("Kind type must be not null");
                if (!type.IsInterface) throw new InvalidOperationException("Kind type must be an interface");
                return type;
            }
        }

        /// <summary>
        /// Order used by <see cref="AmiRelation.Comparer"/>.
        /// </summary>
        public abstract int Order { get; }

        /// <summary>
        /// Get the kind of this attribute.
        /// This must return an interface.
        /// </summary>
        protected abstract Type GetKind { get; }

        /// <summary>
        /// Validates a <paramref name="relation"/> with this Kind.
        /// </summary>
        /// <param name="relation"><see cref="AmiRelation"/> to validate.</param>
        public virtual void Validate(AmiRelation relation)
        {
            if (relation == null) throw new ArgumentNullException("relation");
        }

        /// <summary>
        /// Checks that relation's properties are readonly.
        /// </summary>
        /// <param name="relation">Relation to check.</param>
        protected static void ValidateReadonlyProperties(AmiRelation relation)
        {
            foreach (var property in relation.ReflectedProperties.Where(property => property.CanWrite))
            {
                relation.Add(InfoId.ReadonlyRelation, "Property \"{0}\" should not have a setter", property.CSharpSignature());
            }
        }

        /// <summary>
        /// Validates a <paramref name="type"/> with some relation of this Kind.
        /// </summary>
        /// <param name="type"><see cref="AmiType"/> to validate.</param>
        public virtual void Validate(AmiType type)
        {
            if (type == null) throw new ArgumentNullException("type");
        }

        /// <summary>
        /// Checks that a type has only one relation of a given kind.
        /// </summary>
        /// <param name="type">Type to check</param>
        protected void ValidateSingleRelationPerObject(AmiType type)
        {
            var relations = type.Relations.Where(r => r.Kind != null && r.Kind.Type.Inherits(GetKind));
            if (relations.Skip(1).Any()) type.Add(InfoId.SingleRelationPerObject, "Objects must have only one relation of type {0}", this.KindTypeName());
        }

        /// <summary>
        /// Implements an <see cref="AmiRelation"/>.
        /// The text writer is like:
        /// <code>
        /// namespace N
        /// {
        ///     public class C
        ///     {
        /// </code>
        /// After this method, the text writer will close class and namespace declaration.
        /// This method should not implement 'as' properties.
        /// </summary>
        /// <param name="relation">Relation to implement.</param>
        public abstract string Implement(AmiRelation relation);

        /// <summary>
        /// Implement all properties with a different name than the <paramref name="relation"/>.
        /// </summary>
        /// <param name="code">Where to write the implementation.</param>
        /// <param name="relation">Relation to implement</param>
        protected static string ImplementProperties(AmiRelation relation)
        {
            using (var code = new StringWriter())
            {
                foreach (var asProperty in relation.ReflectedProperties.Where(p => p.Name != relation.Name))
                {
                    var property = asProperty;
                    code.Property(null,
                                  () => code.WriteCSharpQualifiedName(property.PropertyType),
                                  () =>
                                      {
                                          code.WriteCSharpQualifiedName(property.DeclaringType);
                                          code.Write('.');
                                          code.Write(property.Name);
                                      },
                                  () => code.Indent("\t\t\t", () => code.Write("get { return DomainRelation.Domain; } ")));
                }
                return code.ToString();
            }
        }

        /// <summary>
        /// Create the corresponding relation instanciation.
        /// The text writer is like:
        /// <code>
        /// namespace N
        /// {
        ///     public class C
        ///     {
        ///         void CreateRelations()
        ///         {
        /// </code>
        /// </summary>
        /// <param name="relation">Relation to create.</param>
        /// <param name="domain">True if the relation to create is the object's domain relation.</param>
        public abstract string CreateRelation(AmiRelation relation, bool domain);
    }

    /// <summary>
    /// <see cref="RelationKindAttribute"/> extensions.
    /// </summary>
    public static class KindAttributeHelper
    {
        /// <summary>
        /// Write the kind type on the given <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="kindAttribute">What to write.</param>
        public static void WriteKindType(this TextWriter textWriter, RelationKindAttribute kindAttribute)
        {
            const int attributeStringLength = 9; // Length of "Attribute".

            var attributeTypeName = kindAttribute.GetType().Name;
            if (attributeTypeName.Length > attributeStringLength)
            {
                textWriter.Write(attributeTypeName.Substring(0, attributeTypeName.Length - attributeStringLength));
            }
            else
            {
                textWriter.Write(attributeTypeName);
            }
        }

        /// <summary>
        /// Get the kind type name of the given <paramref name="kindAttribute"/>.
        /// </summary>
        /// <param name="kindAttribute"><see cref="RelationKindAttribute"/> to get the name from.</param>
        /// <returns><paramref name="kindAttribute"/> kind type name.</returns>
        public static string KindTypeName(this RelationKindAttribute kindAttribute)
        {
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteKindType(kindAttribute);
                return stringWriter.ToString();
            }
        }
    }
}