#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using Ami.Objects.Core.ReflectionModel;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Abstract <see cref="IClassFactory"/> implementation.
    /// </summary>
    public abstract class AClassFactory : IClassFactory
    {
        /// <summary>
        /// Initializes a new instance of <see cref="AClassFactory"/> with the given <see cref="IObjectModel"/>.
        /// </summary>
        /// <param name="objectModel"><see cref="IClassFactory.ObjectModel"/> value.</param>
        protected AClassFactory(IObjectModel objectModel)
        {
            this.objectModel = objectModel;
        }

        /// <summary>
        /// See: <see cref="IClassFactory.ObjectModel"/>.
        /// </summary>
        private IObjectModel objectModel;

        /// <summary>
        /// <see cref="IObjectModel"/> of this <see cref="IClassFactory"/>.
        /// </summary>
        public IObjectModel ObjectModel
        {
            get { return objectModel; }
        }

        /// <summary>
        /// <see cref="IDomain"/> implementation generated so far.
        /// </summary>
        private Dictionary<Type, Type> domainImplementations;

        /// <summary>
        /// <see cref="IObject"/> implemetations generated so far.
        /// </summary>
        private Dictionary<Type, Type> objectImplementations;

        /// <summary>
        /// Lock object used in <see cref="CreateDomainType{T}"/>.
        /// </summary>
        private readonly object lockODomain = new object();

        /// <summary>
        /// Lock object used in <see cref="CreateObjectType{T}"/>.
        /// </summary>
        private readonly object lockObject = new object();

        /// <summary>
        /// Create a <see cref="IDomain"/> implementation of <typeparamref name="T"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IObjectModel"/>.
        /// </summary>
        /// <typeparam name="T">Type of the domain object to implement (must be an interface).</typeparam>
        /// <returns>An implementation of <typeparamref name="T"/>.</returns>
        public Type CreateDomainType<T>() where T : IDomain
        {
            lock (lockODomain)
            {
                return DoCreate(ref domainImplementations, typeof (T), DoCreateDomainType);
            }
        }

        /// <summary>
        /// Create an implementation of <typeparamref name="T"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IDomain"/>.
        /// </summary>
        /// <typeparam name="T">Type of object to implement (must be an interface).</typeparam>
        /// <returns>An implementation of <typeparamref name="T"/>.</returns>
        public Type CreateObjectType<T>() where T : IObject
        {
            var interfaceType = typeof (T);
            if (interfaceType.Inherits<IDomain>()) throw new InvalidOperationException("Domain classes must be generated with CreateDomainType");
            lock (lockObject)
            {
                return DoCreate(ref objectImplementations, interfaceType, DoCreateObjectType);
            }
        }

        private Type DoCreate(ref Dictionary<Type, Type> implementations, Type interfaceType, Func<AmiType, Type, Type> createMethod)
        {
            if (implementations == null) implementations = new Dictionary<Type, Type>();
            Type implemented;
            if (implementations.TryGetValue(interfaceType, out implemented))
            {
                return implemented;
            }
            var amiType = ObjectModel.Reflector[interfaceType];
            amiType.ThrowErrors();
            implemented = createMethod(amiType, interfaceType);
            implementations.Add(interfaceType, implemented);
            return implemented;
        }

        /// <summary>
        /// Create a <see cref="IDomain"/> implementation of <paramref name="interfaceType"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IObjectModel"/>.
        /// </summary>
        /// <param name="amiType">Ami reflection model of <paramref name="interfaceType"/>.</param>
        /// <param name="interfaceType">Interface to implement.</param>
        /// <returns>An implementation of <paramref name="interfaceType"/>.</returns>
        protected abstract Type DoCreateDomainType(AmiType amiType, Type interfaceType);

        /// <summary>
        /// Create an implementation of <paramref name="interfaceType"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IDomain"/>.
        /// </summary>
        /// <param name="amiType">Ami reflection model of <paramref name="interfaceType"/>.</param>
        /// <param name="interfaceType">Interface to implement.</param>
        /// <returns>An implementation of <paramref name="interfaceType"/>.</returns>
        protected abstract Type DoCreateObjectType(AmiType amiType, Type interfaceType);

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public virtual void Dispose()
        {
            objectModel = null;
            domainImplementations = null;
            objectImplementations = null;
        }
    }
}