#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System.Linq;
using Ami.Objects.Relations;
using Ami.Objects.Streams;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Abstract <see cref="IObject"/> implementation.
    /// </summary>
    /// <typeparam name="TObject">Final object type.</typeparam>
    /// <typeparam name="TOwner">Onwer type of this object.</typeparam>
    /// <typeparam name="TDomain">Domain type of this object.</typeparam>
    public abstract class AObject<TObject, TDomain, TOwner> : IObject<TDomain, TOwner> where TObject : AObject<TObject, TDomain, TOwner> where TOwner : class, IObject where TDomain : class, IDomain
    {
        /// <summary>
        /// <see cref="IObject.Domain"/> relation.
        /// </summary>
        public IDomainRelation<TDomain, TObject> DomainRelation { get; protected set; }

        /// <summary>
        /// <see cref="IDomain"/> of this object.
        /// </summary>
        IDomain IObject.Domain
        {
            get { return Domain; }
        }

        /// <summary>
        /// <see cref="IDomain"/> of this object.
        /// </summary>
        public TDomain Domain
        {
            get { return DomainRelation.Domain; }
        }

        /// <summary>
        /// <see cref="IObject.Owner"/> relation.
        /// </summary>
        public IOwnerRelation<TOwner, TObject> OwnerRelation { get; protected set; }

        /// <summary>
        /// Object's owner, if any.
        /// </summary>
        IObject IObject.Owner
        {
            get { return Owner; }
        }

        /// <summary>
        /// Object's owner, if any.
        /// </summary>
        public TOwner Owner
        {
            get { return OwnerRelation.ObjectOwner; }
        }

        /// <summary>
        /// Enumerates all <see cref="IRelation"/> this <see cref="IObject"/> has.
        /// </summary>
        public IRelation[] Relations { get; private set; }

        /// <summary>
        /// Get all the <see cref="ISerializableRelation"/> that are effectively serialized of this object.
        /// </summary>
        public ISerializableRelation[] SerializedRelations { get; private set; }

        /// <summary>
        /// Get a <see cref="IRelation"/> by its name.
        /// </summary>
        /// <param name="name">Name of the relation to find.</param>
        /// <returns>The first relation with the given name.</returns>
        public IRelation GetRelation(string name)
        {
            return DoGetRelation(name);
        }

        /// <summary>
        /// See <see cref="GetRelation"/>.
        /// </summary>
        /// <param name="name">Name of the relation to find.</param>
        /// <returns>The first relation with the given name.</returns>
        protected abstract IRelation DoGetRelation(string name);

        /// <summary>
        /// Set the content of this <see cref="IObject"/> from the given <paramref name="stream"/>.
        /// </summary>
        /// <param name="stream">Stream from where the content is read.</param>
        public void ReadFrom(IReadableObjectStream stream)
        {
            foreach (var relation in Relations.OfType<ISerializableRelation>())
            {
                relation.ReadFrom(stream);
            }
        }

        /// <summary>
        /// Write the content of this <see cref="IObject"/> to the given <paramref name="stream"/>.
        /// </summary>
        /// <param name="stream">Stream where the content of this object is writen.</param>
        public void WriteTo(IWriteableObjectStream stream)
        {
            foreach (var relation in Relations.OfType<ISerializableRelation>())
            {
                relation.WriteTo(stream);
            }
        }

        /// <summary>
        /// Build the object's relations.
        /// </summary>
        public void BuildRelations()
        {
            Relations = CreateRelations();
            SerializedRelations = Relations.OfType<ISerializableRelation>().Where(r => r.Serialized).ToArray();
        }

        /// <summary>
        /// Creates all the relation of this object.
        /// </summary>
        /// <returns>All relations of this object.</returns>
        protected abstract IRelation[] CreateRelations();

        /// <summary>
        /// Initializes the object's relations, once they are built.
        /// </summary>
        public void InitializeRelations()
        {
            foreach (var relation in Relations) relation.Initialize();
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public virtual void Dispose()
        {
            var relations = Relations;
            ObjectHelper.TryDispose(ref relations);
            Relations = null;
        }
    }
}