#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Abstract implementation of <see cref="IObjectFactory"/>.
    /// </summary>
    public abstract class AObjectFactory : IObjectFactory
    {
        /// <summary>
        /// Initializes a new instance of <see cref="AObjectFactory"/> with the given <paramref name="classFactory"/>.
        /// </summary>
        /// <param name="classFactory"><see cref="IClassFactory"/> used to get implementations.</param>
        protected AObjectFactory(IClassFactory classFactory)
        {
            if (classFactory == null) throw new ArgumentNullException("classFactory");
            var objectModel = classFactory.ObjectModel;
            if (objectModel == null) throw new ArgumentException("The class factory must have an object model.", "classFactory");
            this.classFactory = classFactory;
            this.objectModel = objectModel;
        }

        /// <summary>
        /// See: <see cref="IObjectFactory.ObjectModel"/>.
        /// </summary>
        private IObjectModel objectModel;

        /// <summary>
        /// <see cref="IObjectModel"/> of this <see cref="IObjectFactory"/>.
        /// </summary>
        public IObjectModel ObjectModel
        {
            get { return objectModel; }
        }

        /// <summary>
        /// Class factory used to get implementations.
        /// </summary>
        protected readonly IClassFactory classFactory;

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type of domain to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        public T CreateDomain<T>() where T : IDomain
        {
            var domain = DoCreateDomain<T>();
            domain.BuildRelations();
            domain.InitializeRelations();
            return domain;
        }

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <param name="domain"></param>
        /// <typeparam name="T">Type of object to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        public T CreateObject<T>(IDomain domain) where T : IObject
        {
            if (domain == null) throw new ArgumentNullException("domain");
            var o = DoCreateObject<T>(domain);
            o.BuildRelations();
            o.InitializeRelations();
            return o;
        }

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type of domain to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        protected abstract T DoCreateDomain<T>() where T : IDomain;

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type of object to create, must be an interface.</typeparam>
        /// <param name="domain"><see cref="IDomain"/> of the new <see cref="IObject"/>.</param>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        protected abstract T DoCreateObject<T>(IDomain domain) where T : IObject;

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public void Dispose()
        {
            objectModel = null;
        }
    }
}