#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.CodeDom.Compiler;
using System.Linq;
using System.Text;
using Ami.Objects.Attributes;
using Ami.Objects.Core.ReflectionModel;
using Ami.Objects.Core.Templates;
using Microsoft.CSharp;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Primitive <see cref="IClassFactory"/> implementation.
    /// </summary>
    public class ClassFactory : AClassFactory
    {
        /// <summary>
        /// C# Compiler.
        /// </summary>
        private CSharpCodeProvider compiler;

        /// <summary>
        /// Compiler parameters.
        /// </summary>
        private CompilerParameters parameters;

        /// <summary>
        /// Initailizes a new instance of <see cref="ClassFactory"/> with the given <see cref="IObjectModel"/>.
        /// </summary>
        /// <param name="objectModel"><see cref="IClassFactory"/>'s <see cref="IObjectModel"/>.</param>
        public ClassFactory(IObjectModel objectModel) : base(objectModel)
        {
        }

        private CSharpCodeProvider Compiler
        {
            get { return compiler ?? (compiler = new CSharpCodeProvider()); }
        }

        private CompilerParameters Parameters
        {
            get
            {
                if (parameters == null)
                {
                    parameters = new CompilerParameters {GenerateInMemory = true};
                    parameters.ReferencedAssemblies.Add("System.dll");
                    foreach (var assembly in AppDomain.CurrentDomain.GetAssemblies().Where(a => a.HasAttribute<AmiAssemblyAttribute>()))
                    {
                        var location = assembly.Location;
                        if (!parameters.ReferencedAssemblies.Contains(location)) parameters.ReferencedAssemblies.Add(assembly.Location);
                    }
                    var amiAssembly = GetType().Assembly.Location;
                    if (!parameters.ReferencedAssemblies.Contains(amiAssembly)) parameters.ReferencedAssemblies.Add(amiAssembly);
                }
                return parameters;
            }
        }

        /// <summary>
        /// Create a <see cref="IDomain"/> implementation of <paramref name="interfaceType"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IObjectModel"/>.
        /// </summary>
        /// <param name="amiType">Ami reflection model of <paramref name="interfaceType"/>.</param>
        /// <param name="interfaceType">Interface to implement.</param>
        /// <returns>An implementation of <paramref name="interfaceType"/>.</returns>
        protected override Type DoCreateDomainType(AmiType amiType, Type interfaceType)
        {
            return Compile(amiType, new DomainTemplate(amiType));
        }

        /// <summary>
        /// Create an implementation of <paramref name="interfaceType"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IDomain"/>.
        /// </summary>
        /// <param name="amiType">Ami reflection model of <paramref name="interfaceType"/>.</param>
        /// <param name="interfaceType">Interface to implement.</param>
        /// <returns>An implementation of <paramref name="interfaceType"/>.</returns>
        protected override Type DoCreateObjectType(AmiType amiType, Type interfaceType)
        {
            return Compile(amiType, new ObjectTemplate(amiType));
        }

        private Type Compile(AmiType type, BaseObjectTemplate template)
        {
            var code = template.TransformText();
#if DEBUG
            Console.WriteLine("Generated code for {0}", type.ReflectedType.CSharpQualifiedName());
            Console.WriteLine(code);
#endif
            return CompileCode(code, type);
        }

        private Type CompileCode(string code, AmiType type)
        {
            var interfaceAssembly = Parameters;
            var interfaceAssemblyAdded = false;
            if (!interfaceAssembly.ReferencedAssemblies.Contains(type.ReflectedType.Assembly.Location))
            {
                interfaceAssembly.ReferencedAssemblies.Add(type.ReflectedType.Assembly.Location);
                interfaceAssemblyAdded = true;
            }
            try
            {
                var results = Compiler.CompileAssemblyFromSource(Parameters, code);
                if (results.Errors != null && results.Errors.HasErrors)
                {
                    var report = new StringBuilder();
                    report.AppendLine(results.Errors.ToString());
                    foreach (var error in results.Errors)
                    {
                        report.AppendLine(error.ToString());
                    }
                    throw new InvalidOperationException(report.ToString());
                }
                return results.CompiledAssembly.GetType(type.FullTypeName);
            }
            finally
            {
                if (interfaceAssemblyAdded) interfaceAssembly.ReferencedAssemblies.Remove(type.ReflectedType.Assembly.Location);
            }
        }
    }
}

/*
            using (var code = new StringWriter())
            {
                var fullTypeName = amiType.FullTypeName;
                var typeName = amiType.TypeName;


                var relations = amiType.Relations;
                var domainRelation = amiType.DomainRelation;
                var ownerRelation = amiType.OwnerRelation;
                var domainRelationName = domainRelation.Name;
                var ownerRelationName = ownerRelation.Name;
                var nbRelations = relations.Length;
                if (domainRelation == ownerRelation) ++nbRelations;

                code.Namespace(AmiRuntime, () => code.Class(ImplementationMofidiers, typeName,
                                                            () => code.BaseTypes(() => baseType(code), () => code.WriteCSharpQualifiedName(interfaceType)),
                                                            () =>
                                                                {
                                                                    constructor(code, domainRelationName, ownerRelationName);
                                                                    code.Function(MembersOverrideModifiers,
                                                                                  () => code.WriteCSharpQualifiedName(IRelationArrayType), "CreateRelations",
                                                                                  content: () =>
                                                                                               {
                                                                                                   foreach (var relation in amiType.Relations.Where(r => r != amiType.DomainRelation))
                                                                                                   {
                                                                                                       relation.Kind.CreateRelation(relation);
                                                                                                   }
                                                                                                   code.Variable("relations", () =>
                                                                                                                                  {
                                                                                                                                      code.Write("new ");
                                                                                                                                      code.WriteCSharpQualifiedName(IRelationType);
                                                                                                                                      code.Write("[");
                                                                                                                                      code.Write(nbRelations);
                                                                                                                                      code.Write("]");
                                                                                                                                  });
                                                                                                   for (var i = 0; i < nbRelations; ++i)
                                                                                                   {
                                                                                                       AmiRelation relation;
                                                                                                       switch (i)
                                                                                                       {
                                                                                                           case 0:
                                                                                                               relation = domainRelation;
                                                                                                               break;
                                                                                                           case 1:
                                                                                                               relation = ownerRelation;
                                                                                                               break;
                                                                                                           default:
                                                                                                               relation = relations[domainRelation == ownerRelation ? i - 1 : i];
                                                                                                               break;
                                                                                                       }
                                                                                                       var relationIndex = i;
                                                                                                       code.Assign(() =>
                                                                                                                       {
                                                                                                                           code.Write("relations[");
                                                                                                                           code.Write(relationIndex);
                                                                                                                           code.Write(']');
                                                                                                                       },
                                                                                                                   () =>
                                                                                                                       {
                                                                                                                           code.Write(relation.Name);
                                                                                                                           code.Write("Relation;");
                                                                                                                       });
                                                                                                   }
                                                                                                   code.Indent("\t\t\t", () => code.Write("return relations;"));
                                                                                               });
                                                                    code.Function(MembersOverrideModifiers, () => code.WriteCSharpQualifiedName(IRelationType), 
                                                                        "DoGetRelation", () => code.Parameter(() => code.WriteCSharpQualifiedName(StringType), "name"), 
                                                                        () =>
                                                                            {
                                                                                code.Write("switch (name)");
                                                                                code.Write("{");
                                                                            });
                                                                    foreach (var relation in amiType.Relations)
                                                                    {
                                                                        relation.Kind.Implement(code, ObjectModel, relation);
                                                                    }
                                                                }));
  
  
  
  
  
  
            return CreateType(amiType, interfaceType,
                              code => code.GenericType("Ami.Objects.Core.AObject",
                                                       () => code.Arguments(
                                                           () => code.Write(amiType.TypeName),
                                                           () => code.WriteCSharpQualifiedName(amiType.DomainRelation.RelationType),
                                                           () => code.WriteCSharpQualifiedName(amiType.OwnerRelation.RelationType))),
                              (code, domain, owner) => code.Constructor(ConstructorModifiers, amiType.TypeName,
                                                                        () => code.Parameter(() => code.WriteCSharpQualifiedName(amiType.DomainRelation.RelationType), "domain"),
                                                                        content: () => amiType.DomainRelation.Kind.CreateRelation(amiType.DomainRelation)));
  
  
            return CreateType(amiType, interfaceType,
                              code => code.GenericType("Ami.Objects.Core.ADomain",
                                                       () => code.Arguments(
                                                           () => code.Write(amiType.TypeName),
                                                           () => code.WriteCSharpQualifiedName(amiType.DomainRelation.RelationType),
                                                           () => code.WriteCSharpQualifiedName(amiType.OwnerRelation.RelationType))),
                              (code, domain, owner) => code.Constructor(ConstructorModifiers, amiType.TypeName,
                                                                        () => code.Parameter(() => code.Write("Ami.Objects.IObjectModel"), "objectModel"),
                                                                        () => code.Arguments(() => code.Write("objectModel")), () => amiType.DomainRelation.Kind.CreateRelation(amiType.DomainRelation)));
  
  */