#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Object compiling implementations of <see cref="IObject"/>.
    /// </summary>
    public interface IClassFactory : IDisposable
    {
        /// <summary>
        /// <see cref="IObjectModel"/> of this <see cref="IClassFactory"/>.
        /// </summary>
        IObjectModel ObjectModel { get; }

        /// <summary>
        /// Create a <see cref="IDomain"/> implementation of <typeparamref name="T"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IObjectModel"/>.
        /// </summary>
        /// <typeparam name="T">Type of the domain object to implement (must be an interface).</typeparam>
        /// <returns>An implementation of <typeparamref name="T"/>.</returns>
        Type CreateDomainType<T>() where T : IDomain;

        /// <summary>
        /// Create an implementation of <typeparamref name="T"/>.
        /// The implementation's constructor takes one parameter of type <see cref="IDomain"/>.
        /// </summary>
        /// <typeparam name="T">Type of object to implement (must be an interface).</typeparam>
        /// <returns>An implementation of <typeparamref name="T"/>.</returns>
        Type CreateObjectType<T>() where T : IObject;
    }

    /// <summary>
    /// <see cref="IClassFactory"/> extensions.
    /// </summary>
    public static class ClassFactoryHelper
    {
        /// <summary>
        /// Write all the parameter qualified types and names.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="parameters">What to write.</param>
        public static void WriteParameters(this TextWriter code, IEnumerable<ParameterInfo> parameters)
        {
            code.WriteList(parameters, p =>
                                           {
                                               code.WriteCSharpQualifiedName(p.ParameterType);
                                               code.Write(" ");
                                               code.Write(p.Name);
                                           });
        }


        /// <summary>
        /// Write all the parameter names.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="parameters">What to write.</param>
        public static void WriteParameterNames(this TextWriter code, IEnumerable<ParameterInfo> parameters)
        {
            code.WriteList(parameters, p => code.Write(p.Name));
        }

        /// <summary>
        /// Append a namespace block.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="name">Namespace name.</param>
        /// <param name="content">Content to write in the namespace.</param>
        public static void Namespace(this TextWriter code, string name, Action content)
        {
            code.Write("namespace ");
            code.WriteLine(name);
            code.WriteLine("{");
            content();
            code.WriteLine("}");
        }

        /// <summary>
        /// Append a class block.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="modifiers">Class modifiers (public, private, etc).</param>
        /// <param name="name">Class name.</param>
        /// <param name="baseTypes">Base class and/or interfaces.</param>
        /// <param name="content">Class content.</param>
        /// <param name="indent">Block indentation.</param>
        public static void Class(this TextWriter code, string modifiers, string name, Action baseTypes = null, Action content = null, string indent = "\t")
        {
            code.Indent(indent, () =>
                                    {
                                        code.Write(modifiers);
                                        code.Write(" class ");
                                        code.Write(name);
                                        if (baseTypes != null)
                                        {
                                            code.Write(" : ");
                                            baseTypes();
                                        }
                                    }
                );
            code.Indent(indent, () => code.Write("{"));
            if (content != null) content();
            code.Indent(indent, () => code.Write("}"));
        }

        /// <summary>
        /// Write a generic type name.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="typeName">Generic type name.</param>
        /// <param name="genericParameters">Generic parameters.</param>
        public static void GenericType(this TextWriter code, string typeName, params Action[] genericParameters)
        {
            code.GenericType(typeName, (IEnumerable<Action>) genericParameters);
        }

        /// <summary>
        /// Write a generic type name.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="typeName">Generic type name.</param>
        /// <param name="genericParameters">Generic parameters.</param>
        public static void GenericType(this TextWriter code, string typeName, IEnumerable<Action> genericParameters)
        {
            code.Write(typeName);
            code.Write("<");
            code.WriteList(genericParameters, a => a());
            code.Write(">");
        }

        /// <summary>
        /// Create a list of base types.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="types">What to write.</param>
        public static void BaseTypes(this TextWriter code, params Action[] types)
        {
            code.BaseTypes((IEnumerable<Action>) types);
        }

        /// <summary>
        /// Create a list of base types.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="types">What to write.</param>
        public static void BaseTypes(this TextWriter code, IEnumerable<Action> types)
        {
            code.WriteList(types, a => a());
        }

        /// <summary>
        /// Write a constructor.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="modifiers">Constructor modifiers.</param>
        /// <param name="name">Type name.</param>
        /// <param name="parameters">Parameter definitions.</param>
        /// <param name="parameterNames">Parameter names in base(...)</param>
        /// <param name="content">Constructor content.</param>
        /// <param name="indent">Indentation.</param>
        public static void Constructor(this TextWriter code, string modifiers, string name, Action parameters = null, Action parameterNames = null, Action content = null, string indent = "\t\t")
        {
            code.Indent(indent, () =>
                                    {
                                        code.Write(modifiers);
                                        code.Write(' ');
                                        code.Write(name);
                                        if (parameters != null)
                                        {
                                            code.Quote(parameters, "(", ")");
                                            if (parameterNames != null)
                                            {
                                                code.Write(" : base");
                                                code.Quote(parameterNames, "(", ")");
                                            }
                                        }
                                    }
                );
            code.Indent(indent, () => code.Write("{"));
            if (content != null) content();
            code.Indent(indent, () => code.Write("}"));
        }

        /// <summary>
        /// Write a parameter definition.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="type">Parameter type.</param>
        /// <param name="name">Parameter name.</param>
        public static void Parameter(this TextWriter code, Action type, string name)
        {
            type();
            code.Write(' ');
            code.Write(name);
        }

        /// <summary>
        /// Write a method.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="modifiers">Method modifiers.</param>
        /// <param name="name">Method name.</param>
        /// <param name="parameters">Method parameters</param>
        /// <param name="content">Method content.</param>
        /// <param name="indent">Indentation.</param>
        public static void Method(this TextWriter code, string modifiers, string name, Action parameters = null, Action content = null, string indent = "\t\t")
        {
            code.Function(modifiers, () => code.Write("void"), name, parameters, content, indent);
        }

        /// <summary>
        /// Write a function.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="modifiers">Method modifiers.</param>
        /// <param name="returnType">Functino return type.</param>
        /// <param name="name">Method name.</param>
        /// <param name="parameters">Method parameters</param>
        /// <param name="content">Method content.</param>
        /// <param name="indent">Indentation.</param>
        public static void Function(this TextWriter code, string modifiers, Action returnType, string name, Action parameters = null, Action content = null, string indent = "\t\t")
        {
            code.Indent(indent, () =>
                                    {
                                        code.Write(modifiers);
                                        code.Write(' ');
                                        returnType();
                                        code.Write(' ');
                                        code.Write(name);
                                        code.Write(' ');
                                        if (parameters != null) code.Quote(parameters, "(", ")");
                                        else code.Write("()");
                                    });
            code.Indent(indent, () => code.Write("{"));
            if (content != null) content();
            code.Indent(indent, () => code.Write("}"));
        }

        /// <summary>
        /// Write a variable.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="type">Variable type.</param>
        /// <param name="name">Variable name.</param>
        /// <param name="initialization">Variable initialization.</param>
        /// <param name="indent">Indentation.</param>
        public static void Variable(this TextWriter code, string name, Action initialization, string indent = "\t\t\t")
        {
            code.Indent(indent, () =>
                                    {
                                        code.Write("var ");
                                        code.Write(name);
                                        code.Write(" = ");
                                        initialization();
                                        code.Write(';');
                                    });
        }

        /// <summary>
        /// Write arguments
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="arguments">What to write.</param>
        public static void Arguments(this TextWriter code, IEnumerable<Action> arguments)
        {
            code.WriteList(arguments, a => a());
        }

        /// <summary>
        /// Write arguments
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="arguments">What to write.</param>
        public static void Arguments(this TextWriter code, params Action[] arguments)
        {
            code.WriteList(arguments, a => a());
        }

        /// <summary>
        /// Write arguments
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="arguments">What to write.</param>
        public static void Arguments(this TextWriter code, params string[] arguments)
        {
            code.WriteList(arguments);
        }

        /// <summary>
        /// Write arguments
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="arguments">What to write.</param>
        public static void Arguments(this TextWriter code, IEnumerable<string> arguments)
        {
            code.WriteList(arguments);
        }

        /// <summary>
        /// Write an assignement.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="member">Member assigned.</param>
        /// <param name="value">Value assigned to <paramref name="member"/>.</param>
        /// <param name="ident">Indentation.</param>
        public static void Assign(this TextWriter code, Action member, Action value, string ident = "\t\t\t")
        {
            code.Indent(ident, () =>
                                   {
                                       member();
                                       code.Write(" = ");
                                       value();
                                   });
        }

        public static void Property(this TextWriter code, string modifiers, Action type, Action name, Action getter = null, Action setter = null, string indent = "\t\t")
        {
            if (modifiers != null)
            {
                code.Indent(indent, () => code.Write(modifiers), false);
                type();
            }
            else
            {
                code.Indent(indent, type, false);
            }
            code.Write(' ');
            name();
            code.WriteLine();
            code.Indent(indent, () => code.Write("{"));
            if (getter != null) getter();
            if (setter != null) setter();
            code.Indent(indent, () => code.Write("}"));
        }

        /// <summary>
        /// Indentation.
        /// </summary>
        /// <param name="code">Where to write.</param>
        /// <param name="indentation">Indentation.</param>
        /// <param name="content">Content to write.</param>
        /// <param name="endLine">True to also end the line.</param>
        public static void Indent(this TextWriter code, string indentation, Action content, bool endLine = true)
        {
            code.Write(indentation);
            content();
            if (endLine) code.WriteLine();
        }
    }
}