#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using Ami.Objects.Attributes;
using Ami.Objects.Core.ReflectionModel;

namespace Ami.Objects.Core
{
    /// <summary>
    /// <see cref="IReflector"/> provide implementation information.
    /// </summary>
    public interface IReflector : IDisposable
    {
        /// <summary>
        /// <see cref="IObjectModel"/> of this <see cref="IReflector"/>.
        /// </summary>
        IObjectModel ObjectModel { get; }

        /// <summary>
        /// Get type information on the given <paramref name="type"/>.
        /// </summary>
        /// <param name="type">Reflected type.</param>
        /// <returns>Type information.</returns>
        AmiType this[Type type] { get; }

        /// <summary>
        /// Reflect all type currently loaded.
        /// This iterates through all ami assemblies (see: <see cref="AmiAssemblyAttribute"/>) and all non-generic interfaces inheriting <see cref="IObject"/>.
        /// </summary>
        void ReflectAll();

        /// <summary>
        /// All reflected types so far.
        /// </summary>
        IEnumerable<AmiType> All { get; }
    }

    /// <summary>
    /// <see cref="IReflector"/> extensions.
    /// </summary>
    public static class ReflectorHelper
    {
        /// <summary>
        /// Write information on all loaded types.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="reflector">Reflector.</param>
        public static void WriteAllInfo(this TextWriter textWriter, IReflector reflector)
        {
            textWriter.WriteList(reflector.All, i => textWriter.WriteInfo(i), Environment.NewLine);
        }

        /// <summary>
        /// Check if a given type has <see cref="AbstractAttribute"/>.
        /// </summary>
        /// <typeparam name="T">Type to check.</typeparam>
        /// <returns>True if the type is abstract, otherwise false.</returns>
        public static bool IsAbstract<T>(this IReflector reflector)
        {
            return reflector[typeof (T)].IsAbstract;
        }
    }
}