#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Linq;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Standard <see cref="IObjectFactory"/> implementation.
    /// </summary>
    public class ObjectFactory : AObjectFactory
    {
        /// <summary>
        /// Initializes a new instance of <see cref="ObjectFactory"/> with the given <paramref name="classFactory"/>.
        /// </summary>
        /// <param name="classFactory"><see cref="IClassFactory"/> used to get implementations.</param>
        public ObjectFactory(IClassFactory classFactory) : base(classFactory)
        {
        }

        private static readonly Type[] DomainConstructorArgTypes = new[] {typeof (IObjectModel)};

        protected override T DoCreateDomain<T>()
        {
            var type = classFactory.CreateDomainType<T>();
            if (type == null) throw new InvalidOperationException("Class factory has not returnd an implementation for " + typeof (T).FullName);

            return (T) type.GetConstructor(DomainConstructorArgTypes).Invoke(new object[] {ObjectModel});
        }

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <param name="domain"></param>
        /// <typeparam name="T">Type of object to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        protected override T DoCreateObject<T>(IDomain domain)
        {
            var type = classFactory.CreateObjectType<T>();
            var domainType = domain.GetType();
            if (type == null) throw new InvalidOperationException("Class factory has not returnd an implementation for " + typeof (T).FullName);
            var constructor = type.GetConstructors().FirstOrDefault(c =>
                                                                        {
                                                                            var parameters = c.GetParameters();
                                                                            return parameters.Length == 1 && domainType.Inherits(parameters[0].ParameterType);
                                                                        });
            if (constructor == null) throw new InvalidOperationException("Class factory has not provided a constructor compatible with " + domainType.FullName);
            return (T) constructor.Invoke(new object[] {domain});
        }
    }
}