#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

namespace Ami.Objects.Core
{
    /// <summary>
    /// Primitive <see cref="IObjectModel"/> implementation.
    /// </summary>
    public class ObjectModel : AObjectModel
    {
        /// <summary>
        /// Creates the <see cref="AObjectModel.ClassFactory"/> when required.
        /// </summary>
        /// <returns>A new <see cref="IClassFactory"/> instance.</returns>
        protected override IClassFactory CreateClassFactory()
        {
            return new ClassFactory(this);
        }

        /// <summary>
        /// Creates the <see cref="AObjectModel.ObjectFactory"/> when required.
        /// </summary>
        /// <returns>A new <see cref="IObjectFactory"/> instance.</returns>
        protected override IObjectFactory CreateObjectFactory()
        {
            return new ObjectFactory(ClassFactory);
        }

        /// <summary>
        /// Creates the <see cref="AObjectModel.Reflector"/> when required.
        /// </summary>
        /// <returns>A new <see cref="IReflector"/> instance.</returns>
        protected override IReflector CreateReflector()
        {
            return new Reflector(this);
        }
    }

    /// <summary>
    /// An <see cref="ObjectModel"/> with a single <typeparamref name="TDomain"/> domain object.
    /// </summary>
    /// <typeparam name="TDomain">Domain type.</typeparam>
    public class ObjectModel<TDomain> : ObjectModel where TDomain : class, IDomain
    {
        private readonly object lockDomain = new object();

        private TDomain domain;

        /// <summary>
        /// Single <typeparamref name="TDomain"/> of this <see cref="ObjectModel"/>.
        /// </summary>
        public TDomain Domain
        {
            get
            {
                lock (lockDomain)
                {
                    return domain ?? (domain = ObjectFactory.CreateDomain<TDomain>());
                }
            }
        }

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type of object to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        public T Create<T>() where T : IObject
        {
            return this.Create<T>(Domain);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public override void Dispose()
        {
            base.Dispose();
            ObjectHelper.TryDispose(ref domain);
        }
    }
}