#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Ami.Objects.Core.ReflectionModel
{
    /// <summary>
    /// Base class for <see cref="AmiType"/> and <see cref="AmiProperty"/>.
    /// </summary>
    public abstract class AmiInfo
    {
        /// <summary>
        /// This info's errors.
        /// </summary>
        private List<Info> errors;

        /// <summary>
        /// This info's warnings.
        /// </summary>
        private List<Info> warnings;

        /// <summary>
        /// Sub information contained in this one.
        /// </summary>
        private List<AmiInfo> subInfos;

        /// <summary>
        /// Add a new info as a sub info of this one.
        /// </summary>
        /// <param name="info">Info to add.</param>
        public void Add(AmiInfo info)
        {
            if (info == null) throw new ArgumentNullException("info");
            if (subInfos == null) subInfos = new List<AmiInfo>();
            subInfos.Add(info);
        }

        /// <summary>
        /// Sub informations.
        /// </summary>
        public IEnumerable<AmiInfo> SubInfos
        {
            get { return subInfos ?? Enumerable.Empty<AmiInfo>(); }
        }

        /// <summary>
        /// True if this info has sub information.
        /// </summary>
        public bool HasSubInfo
        {
            get { return subInfos != null && subInfos.Count > 0; }
        }

        /// <summary>
        /// Add a new <see cref="Info"/>.
        /// </summary>
        /// <param name="id">Info id.</param>
        /// <param name="message">Info message.</param>
        public void Add(Enum id, object message)
        {
            var info = new Info(message, this, id);
            List<Info> infos;
            switch (info.Type)
            {
                case InfoType.Error:
                    if (errors == null) errors = new List<Info>();
                    infos = errors;
                    break;
                case InfoType.Warning:
                    if (warnings == null) warnings = new List<Info>();
                    infos = warnings;
                    break;
                default:
                    throw new ArgumentOutOfRangeException("type");
            }
            infos.Add(info);
        }

        /// <summary>
        /// Add an <see cref="Info"/> with the given message.
        /// </summary>
        /// <param name="id">Info id.</param>
        /// <param name="message">Message.</param>
        public void Add(InfoId id, string message)
        {
            Add(id, (object) message);
        }

        /// <summary>
        /// Add an <see cref="Info"/> with the given format and arguments (using <see cref="string.Format(string,object[])"/>).
        /// </summary>
        /// <param name="id">Info id.</param>
        /// <param name="format">String format.</param>
        /// <param name="args">Format arguments.</param>
        public void Add(InfoId id, string format, params object[] args)
        {
            Add(id, (object) string.Format(format, args));
        }

        /// <summary>
        /// All information (errors and warnings).
        /// </summary>
        public IEnumerable<Info> Infos
        {
            get { return Errors.Union(Warnings); }
        }

        /// <summary>
        /// Errors.
        /// </summary>
        public IEnumerable<Info> Errors
        {
            get { return errors ?? Enumerable.Empty<Info>(); }
        }

        /// <summary>
        /// True if this info has some errors.
        /// </summary>
        public bool HasError
        {
            get { return errors != null && errors.Count > 0; }
        }

        /// <summary>
        /// Warnings.
        /// </summary>
        public IEnumerable<Info> Warnings
        {
            get { return warnings ?? Enumerable.Empty<Info>(); }
        }

        /// <summary>
        /// True if this info has some warnings.
        /// </summary>
        public bool HasWarning
        {
            get { return warnings != null && warnings.Count > 0; }
        }

        /// <summary>
        /// All infos, including <see cref="SubInfos"/>.
        /// </summary>
        public IEnumerable<Info> AllInfos
        {
            get { return Infos.Union(SubInfos.SelectMany(i => i.AllInfos)); }
        }

        /// <summary>
        /// All errors, including <see cref="SubInfos"/>.
        /// </summary>
        public IEnumerable<Info> AllErrors
        {
            get { return Errors.Union(SubInfos.SelectMany(i => i.AllErrors)); }
        }

        /// <summary>
        /// All warnings, including <see cref="SubInfos"/>.
        /// </summary>
        public IEnumerable<Info> AllWarnings
        {
            get { return Warnings.Union(SubInfos.SelectMany(i => i.AllWarnings)); }
        }

        /// <summary>
        /// Write this object's source name for error repporting.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        public abstract void WriteSourceNameTo(TextWriter textWriter);

        /// <summary>
        /// Validates this <see cref="AmiInfo"/> and all <see cref="SubInfos"/>.
        /// </summary>
        public void Validate()
        {
            DoValidate();
            if (subInfos != null) foreach (var info in subInfos) info.Validate();
        }

        /// <summary>
        /// Validates only this <see cref="AmiInfo"/>.
        /// </summary>
        protected abstract void DoValidate();

        /// <summary>
        /// Reflects this <see cref="AmiInfo"/>. and all <see cref="SubInfos"/>.
        /// </summary>
        /// <param name="reflector"><see cref="IReflector"/> used.</param>
        public void Reflect(IReflector reflector)
        {
            if (reflector == null) throw new ArgumentOutOfRangeException("reflector");
            DoReflect(reflector);
            if (subInfos != null) foreach (var subInfo in subInfos) subInfo.Reflect(reflector);
            SubInfosReflected();
        }

        /// <summary>
        /// Reflects only this <see cref="AmiInfo"/>.
        /// </summary>
        /// <param name="reflector"><see cref="IReflector"/> used.</param>
        protected abstract void DoReflect(IReflector reflector);

        /// <summary>
        /// Called when all <see cref="SubInfos"/> have been through <see cref="Reflect"/>.
        /// </summary>
        protected virtual void SubInfosReflected()
        {
        }

        /// <summary>
        /// Throw an <see cref="Exception"/> if this type infor has any errors.
        /// </summary>
        public void ThrowErrors()
        {
            var allErrors = AllErrors;
            if (!allErrors.Any()) return;
            string text;
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteAllErrors(this);
                text = stringWriter.ToString();
            }
            throw new Exception(text);
        }
    }

    /// <summary>
    /// <see cref="AmiInfo"/> extensions.
    /// </summary>
    public static class AmiInfoHelper
    {
        /// <summary>
        /// Write the <paramref name="info"/>'s source name on the <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="info">Source name to write.</param>
        public static void WriteSourceName(this TextWriter textWriter, AmiInfo info)
        {
            info.WriteSourceNameTo(textWriter);
        }

        /// <summary>
        /// Write the all <paramref name="info"/>'s error on the given <paramref name="info"/>.
        /// </summary>
        /// <param name="textWriter">Where to write the errors.</param>
        /// <param name="info">Info.</param>
        public static void WriteAllInfos(this TextWriter textWriter, AmiInfo info)
        {
            foreach (var error in info.AllInfos)
            {
                textWriter.WriteInfo(error);
            }
        }

        /// <summary>
        /// Write the all <paramref name="info"/>'s error on the given <paramref name="info"/>.
        /// </summary>
        /// <param name="textWriter">Where to write the errors.</param>
        /// <param name="info">Info.</param>
        public static void WriteAllErrors(this TextWriter textWriter, AmiInfo info)
        {
            foreach (var error in info.AllErrors)
            {
                textWriter.WriteInfo(error);
                textWriter.WriteLine();
            }
        }
    }
}