#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.IO;
using System.Linq;
using System.Reflection;

namespace Ami.Objects.Core.ReflectionModel
{
    /// <summary>
    /// Ami property info.
    /// </summary>
    public class AmiProperty : AmiInfo
    {
        /// <summary>
        /// Initializes a new instance of <see cref="AmiProperty"/> with the given <paramref name="relation"/>.
        /// </summary>
        /// <param name="relation"><see cref="AmiRelation"/> holding this <see cref="AmiProperty"/>.</param>
        /// <param name="name">Name of this <see cref="AmiProperty"/>.</param>
        /// <param name="reflectedProperties">Reflected properties.</param>
        public AmiProperty(AmiRelation relation, string name, PropertyInfo[] reflectedProperties)
        {
            if (relation == null) throw new ArgumentNullException("relation");
            if (name == null) throw new ArgumentNullException("name");
            if (string.IsNullOrWhiteSpace(name)) throw new ArgumentOutOfRangeException("name", "must be non-empty");
            if (reflectedProperties == null) throw new ArgumentNullException("reflectedProperties");
            Relation = relation;
            Name = name;
            ReflectedProperties = reflectedProperties;
        }

        /// <summary>
        /// Reflects only this <see cref="AmiInfo"/>.
        /// </summary>
        /// <param name="reflector"><see cref="IReflector"/> used.</param>
        protected override void DoReflect(IReflector reflector)
        {
        }

        //private static readonly Type AsAttributeType = typeof (AsAttribute);
        //private static readonly Type NotAmiAttributeType = typeof (NotAmiAttribute);

        //private void GetReflectedProperties()
        //{
        //    var listAsNames = new SortedSet<string>();
        //    var countAsNames = -1;
        //    var properties = new HashSet<PropertyInfo>();
        //    var baseTypes = Relation.BaseTypes;
        //    while (countAsNames != listAsNames.Count)
        //    {
        //        countAsNames = listAsNames.Count;

        //        AddReflectedProperties(listAsNames, properties, Relation.ReflectedType);
        //        foreach (var baseType in baseTypes)
        //        {
        //            AddReflectedProperties(listAsNames, properties, baseType);
        //        }
        //    }
        //    ReflectedProperties = properties.ToArray();
        //    AsNames = listAsNames.ToArray();
        //    Kinds = ReflectedProperties.SelectMany(p => p.GetCustomAttributes(false).OfType<KindAttribute>()).ToArray();
        //}

        //private void AddReflectedProperties(ISet<string> listAsNames, ISet<PropertyInfo> properties, Type type)
        //{
        //    var declaredProperties = type.GetProperties(BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly).Where(p => !p.HasAttribute(NotAmiAttributeType) && (p.Name == Name || listAsNames.Contains(p.Name)));
        //    foreach (var property in declaredProperties)
        //    {
        //        foreach (var asAttribute in property.GetCustomAttributes(AsAttributeType, false).Cast<AsAttribute>().Select(a => a.As))
        //        {
        //            listAsNames.Add(asAttribute);
        //        }
        //        properties.Add(property);
        //    }
        //}

        //private void SelectReflectedProperty()
        //{
        //    PropertyType = ReflectedProperties.Select(p => p.PropertyType).CommonType(p => p);

        //    var selectedKind = Kinds.CommonType(k => k.Type);
        //    Kind = selectedKind;

        //    if (AsNames.Length == 1) AsName = AsNames[0];
        //}

        /// <summary>
        /// <see cref="AmiRelation"/>.
        /// </summary>
        public AmiRelation Relation { get; private set; }

        /// <summary>
        /// <see cref="AmiType"/>.
        /// </summary>
        public AmiType Type
        {
            get { return Relation.Type; }
        }

        /// <summary>
        /// Property name.
        /// </summary>
        public string Name { get; private set; }

        ///// <summary>
        ///// All property names this property is following.
        ///// </summary>
        //public string[] AsNames { get; private set; }

        ///// <summary>
        ///// The property this one is following.
        ///// </summary>
        //public string AsName { get; private set; }

        /// <summary>
        /// All CLR properties represented by this <see cref="AmiProperty"/>.
        /// </summary>
        public PropertyInfo[] ReflectedProperties { get; private set; }

        ///// <summary>
        ///// All property kind.
        ///// </summary>
        //public KindAttribute[] Kinds { get; private set; }

        ///// <summary>
        ///// The property kind.
        ///// </summary>
        //public KindAttribute Kind { get; private set; }

        ///// <summary>
        ///// Selected property type.
        ///// </summary>
        //public Type PropertyType { get; private set; }

        /// <summary>
        /// Write this object's source name for error repporting.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        public override void WriteSourceNameTo(TextWriter textWriter)
        {
            Relation.WriteSourceNameTo(textWriter);
            textWriter.Write('.');
            textWriter.Write(Name);
        }

        /// <summary>
        /// Validates only this <see cref="AmiInfo"/>.
        /// </summary>
        protected override void DoValidate()
        {
            //CheckHasPropertyType();
            //CheckHasKinds();
            //CheckHasCommonKind();
            //CheckHasAtMostOneAs();
            CheckHasNoIndexer();
            CheckHasGetter();
        }


        //private void CheckHasAtMostOneAs()
        //{
        //    if (AsNames.Length > 1) Add(InfoType.Error, InfoId.CE_HasAtMostOneAs, "has more than one 'as': {0}", AsName.ToListString());
        //}

        private void CheckHasNoIndexer()
        {
            if (ReflectedProperties == null) return;
            var indexedProperties = ReflectedProperties.Where(p => p.GetIndexParameters().Length > 0);
            if (indexedProperties.Any())
            {
                Add(InfoId.HasNoIndexer, "has reflected properties with indexer: {0}", indexedProperties.ToListString((t, p) => t.Quote(() => t.WriteCSharpSignature(p), "[", "]")));
            }
        }

        private void CheckHasGetter()
        {
            if (ReflectedProperties == null) return;
            var withoutGetterProperties = ReflectedProperties.Where(p => !p.CanRead);
            if (withoutGetterProperties.Any())
            {
                Add(InfoId.HasGetter, "has reflected properties without getter: {0}", withoutGetterProperties.ToListString((t, p) => t.Quote(() => t.WriteCSharpSignature(p), "[", "]")));
            }
        }
    }

    /// <summary>
    /// <see cref="AmiProperty"/> extensions.
    /// </summary>
    public static class AmiPropertyHelper
    {
        /// <summary>
        /// Write <paramref name="info"/> to <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="info">Info to write.</param>
        /// <param name="indent">Indentation.</param>
        public static void WriteInfo(this TextWriter textWriter, AmiProperty info, string indent = "")
        {
            var thisIndent = indent + "\t";
            textWriter.Write(indent);
            textWriter.Write(info.Name);
            textWriter.Write(": ");
            //textWriter.WriteCSharpQualifiedName(info.PropertyType);
            textWriter.WriteLine(" - [");
            //textWriter.Write(thisIndent);
            //textWriter.Write("Kind: ");
            //if (info.Kinds != null && info.Kinds.Length > 0)
            //{
            //    if (info.Kind != null)
            //    {
            //        textWriter.WriteKindType(info.Kind);
            //        textWriter.Write(' ');
            //    }
            //    if (info.Kind == null || info.Kinds.Length > 1)
            //    {
            //        textWriter.Write('(');
            //        textWriter.WriteList(info.Kinds.Where(k => k != info.Kind), (t, k) => t.WriteKindType(k));
            //        textWriter.Write(')');
            //    }
            //}
            //else
            //{
            //    textWriter.Write("()");
            //}
            //textWriter.WriteLine();
            //textWriter.Write(thisIndent);
            //textWriter.Write("Reflected: ");
            //textWriter.WriteList(info.ReflectedProperties, (t, p) => t.Quote(() => t.WriteCSharpSignature(p), "[", "]"));
            //textWriter.WriteLine();
            //if (info.AsNames.Length > 0)
            //{
            //    if (info.AsName != null)
            //    {
            //        textWriter.Write(thisIndent);
            //        textWriter.Write("As: ");
            //        textWriter.Write(info.AsName);
            //        textWriter.WriteLine();
            //    }
            //    else
            //    {
            //        textWriter.Write(thisIndent);
            //        textWriter.Write("As: (");
            //        textWriter.WriteList(info.AsNames);
            //        textWriter.WriteLine(')');
            //    }
            //}
            textWriter.Write(indent);
            textWriter.Write("]");
        }
    }
}