#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using Ami.Objects.Attributes;

namespace Ami.Objects.Core.ReflectionModel
{
    /// <summary>
    /// Ami relation information.
    /// </summary>
    public class AmiRelation : AmiInfo
    {
        /// <summary>
        /// Orders relation.
        /// Domain relation are always the firsts.
        /// Owner relation are always just after the domain relations.
        /// Then properties, attributes and finaly references.
        /// Practically, relations are ordered by <see cref="RelationKindAttribute.Order"/> then <see cref="AmiRelation.Name"/>.
        /// </summary>
        public class Comparer : IComparer<AmiRelation>
        {
            /// <summary>
            /// A comparer instance.
            /// </summary>
            public static readonly Comparer Instance = new Comparer();

            /// <summary>
            /// Compares two relations according to their kind and names.
            /// </summary>
            /// <returns>
            /// A signed integer that indicates the relative values of <paramref name="x"/> and <paramref name="y"/>, as shown in the following table.Value Meaning Less than zero<paramref name="x"/> is less than <paramref name="y"/>.Zero<paramref name="x"/> equals <paramref name="y"/>.Greater than zero<paramref name="x"/> is greater than <paramref name="y"/>.
            /// </returns>
            /// <param name="x">The first object to compare.</param><param name="y">The second object to compare.</param>
            public int Compare(AmiRelation x, AmiRelation y)
            {
                if (x == null)
                {
                    if (y == null) return 0;
                    return 1;
                }
                if (y == null) return -1;
                if (ReferenceEquals(x, y)) return 0;
                var xKind = x.Kind;
                var yKind = y.Kind;
                if (xKind == null)
                {
                    if (yKind == null) return 0;
                    return 1;
                }
                if (yKind == null) return -1;
                var xOrder = xKind.Order;
                var yOrder = yKind.Order;
                if (xOrder == yOrder)
                {
                    var xName = x.Name;
                    var yName = y.Name;
                    if (xName == null)
                    {
                        if (yName == null) return 0;
                        return 1;
                    }
                    if (yName == null) return -1;
                    return xName.CompareTo(yName);
                }
                return xOrder < yOrder ? -1 : 1;
            }
        }

        /// <summary>
        /// Initializes a new instance of <see cref="AmiRelation"/> which belongs to <paramref name="type"/> with <paramref name="names"/>, <paramref name="kinds"/> and <paramref name="reflectedProperties"/>.
        /// </summary>
        /// <param name="type">Type this relation belongs to.</param>
        /// <param name="names">Names of this relation.</param>
        /// <param name="kinds">Kinds of this relation.</param>
        /// <param name="reflectedProperties">Property infos belonging to this relation.</param>
        public AmiRelation(AmiType type, string[] names, RelationKindAttribute[] kinds, PropertyInfo[] reflectedProperties)
        {
            if (type == null) throw new ArgumentNullException("type");
            if (kinds == null) throw new ArgumentNullException("kinds");
            if (names == null) throw new ArgumentNullException("names");
            if (reflectedProperties == null) throw new ArgumentNullException("reflectedProperties");

            if (names.Length == 0) throw new ArgumentOutOfRangeException("names", "Names must have at least one value");
            if (reflectedProperties.Length == 0) throw new ArgumentOutOfRangeException("reflectedProperties", "Reflected properties must have at least one value");

            Type = type;
            Names = names;
            Kinds = kinds;
            ReflectedProperties = reflectedProperties;
            foreach (var name in names)
            {
                var pName = name;
                Add(new AmiProperty(this, name, reflectedProperties.Where(p => p.Name == pName).ToArray()));
            }
        }

        /// <summary>
        /// Type having this relation.
        /// </summary>
        public AmiType Type { get; private set; }

        /// <summary>
        /// All the name this relation has.
        /// </summary>
        public string[] Names { get; private set; }

        /// <summary>
        /// Selected name.
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// All the kind this relation has.
        /// </summary>
        public RelationKindAttribute[] Kinds { get; private set; }

        /// <summary>
        /// Selected kind.
        /// </summary>
        public RelationKindAttribute Kind { get; private set; }

        /// <summary>
        /// All the properties of this relation.
        /// </summary>
        public PropertyInfo[] ReflectedProperties { get; private set; }

        /// <summary>
        /// Properties of this relation.
        /// </summary>
        public IEnumerable<AmiProperty> Properties
        {
            get { return Infos.OfType<AmiProperty>(); }
        }

        /// <summary>
        /// <see cref="AmiRelation"/> value type.
        /// </summary>
        public Type RelationType { get; private set; }

        /// <summary>
        /// Write this object's source name for error repporting.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        public override void WriteSourceNameTo(TextWriter textWriter)
        {
            Type.WriteSourceNameTo(textWriter);
            textWriter.Write('.');
            textWriter.Write(Name);
        }

        /// <summary>
        /// Reflects only this <see cref="AmiInfo"/>.
        /// </summary>
        /// <param name="reflector"><see cref="IReflector"/> used.</param>
        protected override void DoReflect(IReflector reflector)
        {
            RelationType = ReflectedProperties.Select(p => p.PropertyType).CommonType(p => p);
            Kind = Kinds.CommonType(k => k.Type);
            if (Names.Length == 1) Name = Names[0];
            else if (Kind != null)
            {
                var kindProperties = ReflectedProperties.Where(p => p.GetCustomAttributes(false).OfType<RelationKindAttribute>().Contains(Kind));
                Name = kindProperties.Select(p => p.Name).OneOrDefault() ?? ReflectedProperties.Where(p => p.HasAttribute<RelationNameAttribute>()).Select(p => p.Name).OneOrDefault();
            }
        }

        /// <summary>
        /// Validates only this <see cref="AmiInfo"/>.
        /// </summary>
        protected override void DoValidate()
        {
            CheckHasRelationType();
            CheckHasKinds();
            CheckHasCommonKind();
            CheckHasName();
            if (Kind != null)
            {
                Kind.Validate(this);
                Kind.Validate(Type);
            }
        }

        private void CheckHasRelationType()
        {
            if (RelationType == null) Add(InfoId.HasCommonType, "has no common property type between: {0}", ReflectedProperties.Select(p => p.PropertyType).Distinct().ToListString((t, p) => t.WriteCSharpQualifiedName(p)));
        }

        private void CheckHasKinds()
        {
            if (Kinds == null || Kinds.Length == 0) Add(InfoId.HasKinds, "has no kind");
        }

        private void CheckHasCommonKind()
        {
            if (Kinds == null || Kinds.Length == 0) return;
            if (Kind == null) Add(InfoId.HasCommonKind, "has no common kind between: {0}", Kinds.ToListString((t, p) => t.WriteKindType(p)));
        }

        private void CheckHasName()
        {
            if (Kind != null && Name == null)
            {
                Add(InfoId.HasName, "cannot choose between relation names: {0}", Names.ToListString());
            }
        }
    }

    /// <summary>
    /// <see cref="AmiRelation"/> extensions.
    /// </summary>
    public static class AmiRelationHelper
    {
        /// <summary>
        /// Write <paramref name="info"/> to <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="info">Info to write.</param>
        /// <param name="indent">Indentation.</param>
        public static void WriteInfo(this TextWriter textWriter, AmiRelation info, string indent = "")
        {
            var thisIndent = indent + "\t";
            textWriter.Write(indent);
            textWriter.Write(info.Name ?? "<no-name>");
            textWriter.Write(": ");
            textWriter.WriteCSharpQualifiedName(info.RelationType);
            textWriter.WriteLine(" - [");
            textWriter.Write(thisIndent);
            textWriter.Write("Kind: ");
            if (info.Kinds != null && info.Kinds.Length > 0)
            {
                if (info.Kind != null)
                {
                    textWriter.WriteKindType(info.Kind);
                    textWriter.Write(' ');
                }
                if (info.Kind == null || info.Kinds.Length > 1)
                {
                    textWriter.Write('(');
                    textWriter.WriteList(info.Kinds.Where(k => k != info.Kind), k => textWriter.WriteKindType(k));
                    textWriter.Write(')');
                }
            }
            else
            {
                textWriter.Write("()");
            }
            textWriter.WriteLine();
            textWriter.Write(thisIndent);
            textWriter.Write("Reflected: ");
            textWriter.WriteList(info.ReflectedProperties, p => textWriter.Quote(() => textWriter.WriteCSharpSignature(p), "[", "]"));
            textWriter.WriteLine();
            if (info.Names.Length > 0)
            {
                textWriter.Write(thisIndent);
                textWriter.Write("As: ");
                if (info.Name != null)
                {
                    textWriter.Write(info.Name);
                    textWriter.Write(" ");
                }
                textWriter.Quote(() => textWriter.WriteList(info.Names.Where(n => n != info.Name)), "(", ")");
                textWriter.WriteLine();
            }
            textWriter.Write(indent);
            textWriter.Write("]");
        }
    }
}