﻿#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using Ami.Objects.Attributes;

namespace Ami.Objects.Core.ReflectionModel
{
    /// <summary>
    /// Ami type info.
    /// </summary>
    public class AmiType : AmiInfo
    {
        /// <summary>
        /// Initializes a new instance of <see cref="AmiType"/> for the given <paramref name="reflectedType"/>.
        /// </summary>
        /// <param name="reflectedType">Reflected type by this <see cref="AmiType"/>.</param>
        public AmiType(Type reflectedType)
        {
            if (reflectedType == null) throw new ArgumentOutOfRangeException("reflectedType");
            ReflectedType = reflectedType;

            BaseTypes = ReflectedType.GetInterfaces().Where(i => i.Inherits(IObjectType)).ToArray();

            var clrProperties = BaseTypes.SelectMany(t => t.GetProperties(BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly))
                .Union(ReflectedType.GetProperties(BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly))
                .Where(p => !p.HasAttribute(NotAmiAttributeType)).ToArray();

            var propertyNames = clrProperties.Select(p => p.Name).Distinct().ToArray();

            var relationGroups = new List<HashSet<string>>();

            foreach (var propertyName in propertyNames)
            {
                var pName = propertyName;
                if (relationGroups.Any(g => g.Contains(pName))) continue;
                var properties = clrProperties.Where(p => p.Name == pName);
                var asAttributes = properties.SelectMany(p => p.GetCustomAttributes(false).OfType<AsAttribute>());
                var groupFound = false;
                foreach (var asAttribute in asAttributes)
                {
                    var asName = asAttribute.As;
                    if (pName == asName) continue;
                    foreach (var group in relationGroups)
                    {
                        if (group.Contains(pName)) continue;
                        if (group.Contains(asName))
                        {
                            group.Add(pName);
                            groupFound = true;
                        }
                    }
                }
                if (!groupFound)
                {
                    var group = new HashSet<string> {pName};
                    foreach (var asAttribute in asAttributes) group.Add(asAttribute.As);
                    relationGroups.Add(group);
                }
            }

            foreach (var relationGroup in relationGroups)
            {
                var group = relationGroup.ToArray();
                var properties = clrProperties.Where(p => group.Contains(p.Name)).ToArray();
                Add(new AmiRelation(this, group, properties.SelectMany(p => p.GetCustomAttributes(false).OfType<RelationKindAttribute>()).ToArray(), properties));
            }
        }

        private static readonly Type IObjectType = typeof (IObject);
        private static readonly Type IDomainType = typeof (IDomain);
        private static readonly Type NotAmiAttributeType = typeof (NotAmiAttribute);
        private static readonly Type AbstractAttributeType = typeof (AbstractAttribute);
        private static readonly Type DomainAttributeKindType = typeof (DomainAttribute.IKind);
        private static readonly Type OwnerAttributeKindType = typeof (OwnerAttribute.IKind);

        /// <summary>
        /// Reflects only this <see cref="AmiInfo"/>.
        /// </summary>
        /// <param name="reflector"><see cref="IReflector"/> used.</param>
        protected override void DoReflect(IReflector reflector)
        {
            var helperName = ReflectedType.ChangeCSharpQualifiedName((ReflectedType.Name.Length > 1 ? ReflectedType.Name.Substring(1) : ReflectedType.Name) + "Helper");
            HelperType = ReflectedType.Assembly.FindTypeByCSharpQualifiedName(helperName);
            IsAbstract = ReflectedType.HasAttribute(AbstractAttributeType);
            IsDomainType = ReflectedType.Inherits(IDomainType);
            TypeName = ReflectedType.Name.Substring(1);
            FullTypeName = "Ami.Runtime." + TypeName;
        }

        /// <summary>
        /// Called when all <see cref="AmiInfo.SubInfos"/> have been through <see cref="AmiInfo.Reflect"/>.
        /// </summary>
        protected override void SubInfosReflected()
        {
            base.SubInfosReflected();
            var relations = SubInfos.OfType<AmiRelation>().ToList();
            relations.Sort(AmiRelation.Comparer.Instance);
            Relations = relations.ToArray();
            if (ReflectedType.IsInterface && ReflectedType.Inherits(IObjectType)) // a class or non IObject will have no relations, Single will throw an exception.
            {
                DomainRelation = Relations.Where(r => r.Kind != null && r.Kind.Type == DomainAttributeKindType).FirstOrDefault() ?? Relations.Where(r => r.Kind != null && r.Kind.Type.Inherits(DomainAttributeKindType)).Single();
                OwnerRelation = Relations.Where(r => r.Kind != null && r.Kind.Type == OwnerAttributeKindType).FirstOrDefault() ?? Relations.Where(r => r.Kind != null && r.Kind.Type.Inherits(OwnerAttributeKindType)).Single();
            }
        }

        /// <summary>
        /// The <see cref="Type"/> this information in on.
        /// </summary>
        public Type ReflectedType { get; private set; }

        /// <summary>
        /// Helper type, if any.
        /// </summary>
        public Type HelperType { get; private set; }

        /// <summary>
        /// True if the <see cref="ReflectedType"/> is abstract.
        /// </summary>
        public bool IsAbstract { get; private set; }

        /// <summary>
        /// True if the <see cref="ReflectedType"/> is a <see cref="IDomain"/>.
        /// </summary>
        public bool IsDomainType { get; private set; }

        /// <summary>
        /// All base type information (interfaces) of <see cref="ReflectedType"/>.
        /// </summary>
        public Type[] BaseTypes { get; private set; }

        /// <summary>
        /// Get all relations of this type.
        /// </summary>
        public AmiRelation[] Relations { get; private set; }

        /// <summary>
        /// Domain relation.
        /// </summary>
        public AmiRelation DomainRelation { get; private set; }

        /// <summary>
        /// Onwer relation.
        /// </summary>
        public AmiRelation OwnerRelation { get; private set; }

        /// <summary>
        /// Full implementation type name.
        /// </summary>
        public string FullTypeName { get; private set; }

        /// <summary>
        /// Implementation type name.
        /// </summary>
        public string TypeName { get; private set; }

        /// <summary>
        /// Write this object's source name for error repporting.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        public override void WriteSourceNameTo(TextWriter textWriter)
        {
            textWriter.WriteCSharpQualifiedName(ReflectedType);
        }

        /// <summary>
        /// Validates only this <see cref="AmiInfo"/>.
        /// </summary>
        protected override void DoValidate()
        {
            CheckIsIObject();
            CheckIsInterface();
            CheckNameStartsWithI();
        }

        private void CheckIsIObject()
        {
            if (ReflectedType != null && !ReflectedType.Inherits<IObject>()) Add(InfoId.IsIObject, "Type is not an IObject");
        }

        private void CheckIsInterface()
        {
            if (ReflectedType != null && !ReflectedType.IsInterface) Add(InfoId.IsInterface, "Type is not an interface");
        }

        private void CheckNameStartsWithI()
        {
            if (ReflectedType != null && !ReflectedType.Name.StartsWith("I")) Add(InfoId.NameStartsWithI, "Type name doesn't starts with an I");
        }
    }

    public static class AmiTypeHelper
    {
        /// <summary>
        /// Write <paramref name="type"/> to <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="type">Info to write.</param>
        /// <param name="indent">Indentation.</param>
        public static void WriteInfo(this TextWriter textWriter, AmiType type, string indent = "")
        {
            if (type == null)
            {
                textWriter.Write(indent);
                textWriter.Write("<null-type>");
                return;
            }
            textWriter.Write(indent);
            textWriter.Write(type.ReflectedType.CSharpQualifiedName());
            textWriter.WriteLine(" - [");
            var thisIndent = indent + '\t';
            if (type.HelperType != null)
            {
                textWriter.Write(thisIndent);
                textWriter.Write("Helper type: ");
                textWriter.WriteCSharpQualifiedName(type.HelperType);
                textWriter.WriteLine();
            }
            var directBaseTypes = type.BaseTypes;
            if (directBaseTypes != null && directBaseTypes.Length > 0)
            {
                textWriter.Write(thisIndent);
                textWriter.Write("Base types: ");
                textWriter.WriteList(directBaseTypes, i => textWriter.WriteCSharpQualifiedName(i));
                textWriter.WriteLine();
            }

            var relations = type.Relations;
            var nextIndent = thisIndent + '\t';
            if (relations != null && relations.Any())
            {
                textWriter.Write(thisIndent);
                textWriter.WriteLine("Relations: [");
                foreach (var relation in relations)
                {
                    textWriter.WriteInfo(relation, nextIndent);
                    textWriter.WriteLine();
                }
                textWriter.Write(thisIndent);
                textWriter.WriteLine("]");
            }
            textWriter.WriteLine("]");
        }
    }
}