#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.IO;

namespace Ami.Objects.Core.ReflectionModel
{
    /// <summary>
    /// Structure recording info message, source and type.
    /// </summary>
    public struct Info
    {
        /// <summary>
        /// Info message.
        /// </summary>
        public readonly object Message;

        /// <summary>
        /// Info source.
        /// </summary>
        public readonly AmiInfo Source;

        /// <summary>
        /// Information type.
        /// </summary>
        public readonly InfoType Type;

        /// <summary>
        /// Infomration id.
        /// </summary>
        public readonly Enum Id;

        /// <summary>
        /// Initializes an error.
        /// </summary>
        /// <param name="message">Info message.</param>
        /// <param name="source">Info source.</param>
        /// <param name="id">Info id.</param>
        public Info(object message, AmiInfo source, Enum id)
        {
            if (message == null) throw new ArgumentNullException("message");
            if (source == null) throw new ArgumentNullException("source");
            if (id == null) throw new ArgumentNullException("id");

            Message = message;
            Source = source;
            Id = id;
            Type = ErrorAttribute.IsError(id) ? InfoType.Error : InfoType.Warning;
        }

        /// <summary>
        /// True when this info is an error.
        /// </summary>
        public bool IsError
        {
            get { return Type == InfoType.Error; }
        }

        /// <summary>
        /// True when this information is a warning.
        /// </summary>
        public bool IsWarning
        {
            get { return Type == InfoType.Warning; }
        }

        /// <summary>
        /// Format: "Source: Message"
        /// </summary>
        /// <returns>Info string presentation.</returns>
        public override string ToString()
        {
            return string.Format("[{0}] {1}: {2}", Id, Source, Message);
        }
    }

    /// <summary>
    /// <see cref="Info"/> extensions.
    /// </summary>
    public static class InfoHelper
    {
        /// <summary>
        /// Write <paramref name="info"/> to the <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="info">What to write.</param>
        public static void WriteInfo(this TextWriter textWriter, Info info)
        {
            textWriter.Write('[');
            textWriter.Write(info.Id);
            textWriter.Write("] ");
            textWriter.WriteSourceName(info.Source);
            textWriter.Write(": ");
            textWriter.Write(info.Message);
        }
    }
}