#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using Ami.Objects.Attributes;
using Ami.Objects.Core.ReflectionModel;

namespace Ami.Objects.Core
{
    /// <summary>
    /// Implementation of <see cref="IReflector"/>.
    /// </summary>
    public class Reflector : IReflector
    {
        private readonly object lockObject = new object();

        /// <summary>
        /// Holds reflection results.
        /// </summary>
        private Dictionary<Type, AmiType> infos;

        /// <summary>
        /// See : <see cref="ObjectModel"/>.
        /// </summary>
        private IObjectModel objectModel;

        /// <summary>
        /// Initailizes a new instance of <see cref="Reflector"/> with the given <paramref name="objectModel"/>.
        /// </summary>
        /// <param name="objectModel"><see cref="IReflector"/>'s object model.</param>
        public Reflector(IObjectModel objectModel)
        {
            this.objectModel = objectModel;
        }

        /// <summary>
        /// <see cref="IObjectModel"/> of this <see cref="IReflector"/>.
        /// </summary>
        public IObjectModel ObjectModel
        {
            get { return objectModel; }
        }

        /// <summary>
        /// Get type information on the given <paramref name="type"/>.
        /// </summary>
        /// <param name="type">Reflected type.</param>
        /// <returns>Type information.</returns>
        public AmiType this[Type type]
        {
            get { return GetInfo(type); }
        }

        /// <summary>
        /// Reflect all type currently loaded.
        /// This iterates through all ami assemblies (see: <see cref="AmiAssemblyAttribute"/>) and all non-generic interfaces inheriting <see cref="IObject"/>.
        /// </summary>
        public void ReflectAll()
        {
            var amiAssemblyAttrType = typeof (AmiAssemblyAttribute);
            var iobjectType = typeof (IObject);

            foreach (var loadedAssembly in AppDomain.CurrentDomain.GetAssemblies().Where(a => a.HasAttribute(amiAssemblyAttrType)))
            {
                foreach (var type in loadedAssembly.GetTypes().Where(t => t.IsInterface && !t.IsGenericType && t.Inherits(iobjectType)))
                {
                    GetInfo(type);
                }
            }
        }

        /// <summary>
        /// All reflected types so far.
        /// </summary>
        public IEnumerable<AmiType> All
        {
            get { return infos != null ? infos.Select(p => p.Value) : Enumerable.Empty<AmiType>(); }
        }

        /// <summary>
        /// Get <see cref="AmiType"/> for the given <paramref name="type"/>.
        /// </summary>
        /// <param name="type">Type to get reflection.</param>
        /// <returns>Reflection info.</returns>
        protected AmiType GetInfo(Type type)
        {
            if (type == null) throw new ArgumentNullException("type");
            lock (lockObject)
            {
                if (infos == null) infos = new Dictionary<Type, AmiType>();
                AmiType info;
                if (infos.TryGetValue(type, out info))
                {
                    return info;
                }
                info = new AmiType(type);
                infos.Add(type, info);
                info.Reflect(this);
                info.Validate();
                return info;
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public virtual void Dispose()
        {
            objectModel = null;
            infos = null;
        }
    }
}