﻿#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Ami.Objects.Attributes;
using Ami.Objects.Core;
using Ami.Objects.Relations;
using Ami.Objects.Streams;

namespace Ami.Objects
{
    /// <summary>
    /// Common interface of all "Ami" objects.
    /// </summary>
    [Abstract]
    public interface IObject : IDisposable
    {
        /// <summary>
        /// Build the object's relations.
        /// </summary>
        [NotAmi]
        void BuildRelations();

        /// <summary>
        /// Initializes the object's relations, once they are built (see <see cref="BuildRelations"/>).
        /// Calling this method will reset all relations to their original values, just like the object has just been created.
        /// </summary>
        [NotAmi]
        void InitializeRelations();

        /// <summary>
        /// <see cref="IDomain"/> of this object.
        /// </summary>
        [Domain]
        IDomain Domain { get; }

        /// <summary>
        /// Object's owner, if any.
        /// </summary>
        [Owner]
        IObject Owner { get; }

        /// <summary>
        /// Get a <see cref="IRelation"/> by its name.
        /// </summary>
        /// <param name="name">Name of the relation to find.</param>
        /// <returns>The first relation with the given name.</returns>
        [NotAmi]
        IRelation GetRelation(string name);

        /// <summary>
        /// Get all the relations of this object.
        /// </summary>
        [NotAmi]
        IRelation[] Relations { get; }

        /// <summary>
        /// Get all the <see cref="ISerializableRelation"/> that are effectively serialized of this object.
        /// </summary>
        [NotAmi]
        ISerializableRelation[] SerializedRelations { get; }

        /// <summary>
        /// Set the content of this <see cref="IObject"/> from the given <paramref name="stream"/>.
        /// </summary>
        /// <param name="stream">Stream from where the content is read.</param>
        [NotAmi]
        void ReadFrom(IReadableObjectStream stream);

        /// <summary>
        /// Write the content of this <see cref="IObject"/> to the given <paramref name="stream"/>.
        /// </summary>
        /// <param name="stream">Stream where the content of this object is writen.</param>
        [NotAmi]
        void WriteTo(IWriteableObjectStream stream);
    }

    /// <summary>
    /// Typed <see cref="IObject"/>.
    /// </summary>
    /// <typeparam name="TDomain">Domain type.</typeparam>
    [Abstract]
    public interface IObject<out TDomain> : IObject where TDomain : IDomain
    {
        /// <summary>
        /// <see cref="IDomain"/> of this object.
        /// </summary>
        [Domain]
        new TDomain Domain { get; }
    }

    /// <summary>
    /// Typed common interface to all Ami objects.
    /// </summary>
    /// <typeparam name="TDomain">Domain type.</typeparam>
    /// <typeparam name="TOnwer">Onwer type.</typeparam>
    [Abstract]
    public interface IObject<out TDomain, out TOnwer> : IObject<TDomain> where TDomain : IDomain where TOnwer : IObject
    {
        /// <summary>
        /// Object's owner, if any.
        /// </summary>
        [Owner]
        new TOnwer Owner { get; }
    }

    /// <summary>
    /// <see cref="IObject"/> extensions and other common extension methods.
    /// </summary>
    public static class ObjectHelper
    {
        /// <summary>
        /// Dispose <paramref name="toDispose"/> if it is not null and then set <paramref name="toDispose"/> to null.
        /// </summary>
        /// <typeparam name="T">Type of object to dispose.</typeparam>
        /// <param name="toDispose">Object to dispose.</param>
        public static void TryDispose<T>(ref T toDispose) where T : class, IDisposable
        {
            if (toDispose == null) return;
            toDispose.Dispose();
            toDispose = null;
        }

        /// <summary>
        /// Dispose <paramref name="toDispose"/> if it is not null and then set <paramref name="toDispose"/> to null.
        /// </summary>
        /// <typeparam name="T">Type of object to dispose.</typeparam>
        /// <param name="toDispose">Object array to dispose.</param>
        public static void TryDispose<T>(ref T[] toDispose) where T : class, IDisposable
        {
            if (toDispose == null) return;
            for (var i = 0; i < toDispose.Length; ++i)
            {
                var item = toDispose[i];
                TryDispose(ref item);
            }
            toDispose = null;
        }

        /// <summary>
        /// Checks an <paramref name="enumerable"/> has one and only one element.
        /// </summary>
        /// <typeparam name="T"><paramref name="enumerable"/>' element type.</typeparam>
        /// <param name="enumerable">Enumerable to check.</param>
        /// <returns>True if <paramref name="enumerable"/> has one and only one element, otherwise false.</returns>
        public static bool HasOne<T>(this IEnumerable<T> enumerable)
        {
            return enumerable.Any() && (!enumerable.Skip(1).Any());
        }

        /// <summary>
        /// Get the single and only elment if the given <paramref name="enumerable"/> has one and only one element, otherwise returns null.
        /// </summary>
        /// <typeparam name="T"><paramref name="enumerable"/>' element type.</typeparam>
        /// <param name="enumerable">Enumerable to get the element from.</param>
        /// <returns>The first element of <paramref name="enumerable"/> or null if their is zero or more than one element.</returns>
        public static T OneOrDefault<T>(this IEnumerable<T> enumerable)
        {
            var firstOrDefault = enumerable.FirstOrDefault();
            if (enumerable.Any() && (!enumerable.Skip(1).Any())) return firstOrDefault;
            return default(T);
        }

        /// <summary>
        /// Get the object's <see cref="IObjectModel"/>.
        /// </summary>
        /// <param name="o"><see cref="IObject"/></param>
        /// <returns><paramref name="o"/>'s <see cref="IObjectModel"/>.</returns>
        public static IObjectModel ObjectModel(this IObject o)
        {
            return o.Domain.ObjectModel;
        }

        /// <summary>
        /// Get the object's <see cref="IClassFactory"/>.
        /// </summary>
        /// <param name="o"><see cref="IObject"/>.</param>
        /// <returns><paramref name="o"/>'s <see cref="IClassFactory"/></returns>
        public static IClassFactory ClassFactory(this IObject o)
        {
            return o.ObjectModel().ClassFactory;
        }

        /// <summary>
        /// Get the object's <see cref="IObjectFactory"/>.
        /// </summary>
        /// <param name="o"><see cref="IObject"/>.</param>
        /// <returns><paramref name="o"/>'s <see cref="IObjectFactory"/>.</returns>
        public static IObjectFactory ObjectFactory(this IObject o)
        {
            return o.ObjectModel().ObjectFactory;
        }

        /// <summary>
        /// Write the given enumerable <paramref name="list"/> to the <paramref name="textWriter"/> applying the <paramref name="write"/> action to each elements.
        /// </summary>
        /// <typeparam name="T">Enumerable item type.</typeparam>
        /// <param name="textWriter">Where the list is  written to.</param>
        /// <param name="list">Enumerable to write.</param>
        /// <param name="write">Write action.</param>
        /// <param name="separator">Separator between elements.</param>
        /// <param name="startWithSeparator">True to always start with a separator.</param>
        /// <param name="finishWithSeparator">True to always finish with a separator.</param>
        public static void WriteList<T>(this TextWriter textWriter, IEnumerable<T> list, Action<T> write, string separator = ", ", bool startWithSeparator = false, bool finishWithSeparator = false)
        {
            if (startWithSeparator) textWriter.Write(separator);
            if (list != null)
            {
                var i = 0;
                foreach (var item in list)
                {
                    if (i > 0) textWriter.Write(separator);
                    write(item);
                    ++i;
                }
                if (finishWithSeparator && i > 0) textWriter.Write(separator);
            }
        }

        /// <summary>
        /// Write the given enumerable <paramref name="list"/> to the <paramref name="textWriter"/> applying the <typeparamref name="T"/> <see cref="Object.ToString()"/> method to each elements.
        /// </summary>
        /// <typeparam name="T">Enumerable item type.</typeparam>
        /// <param name="textWriter">Where the list is  written to.</param>
        /// <param name="list">Enumerable to write.</param>
        /// <param name="separator">Separator between elements.</param>
        /// <param name="startWithSeparator">True to always start with a separator.</param>
        /// <param name="finishWithSeparator">True to always finish with a separator.</param>
        public static void WriteList<T>(this TextWriter textWriter, IEnumerable<T> list, string separator = ", ", bool startWithSeparator = false, bool finishWithSeparator = false)
        {
            textWriter.WriteList(list, i => textWriter.Write(i.ToString()), separator, startWithSeparator, finishWithSeparator);
        }

        /// <summary>
        /// Write the given enumerable <paramref name="list"/> to a string applying the <paramref name="write"/> action to each elements.
        /// </summary>
        /// <typeparam name="T">Enumerable item type.</typeparam>
        /// <param name="list">Enumerable to write.</param>
        /// <param name="write">Write action.</param>
        /// <param name="separator">Separator between elements.</param>
        /// <param name="startWithSeparator">True to always start with a separator.</param>
        /// <param name="finishWithSeparator">True to always finish with a separator.</param>
        public static string ToListString<T>(this IEnumerable<T> list, Action<TextWriter, T> write, string separator = ", ", bool startWithSeparator = false, bool finishWithSeparator = false)
        {
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteList(list, i => write(stringWriter, i), separator, startWithSeparator, finishWithSeparator);
                return stringWriter.ToString();
            }
        }

        /// <summary>
        /// Write the given enumerable <paramref name="list"/> to a string using the item's to string on each elements.
        /// </summary>
        /// <typeparam name="T">Enumerable item type.</typeparam>
        /// <param name="list">Enumerable to write.</param>
        /// <param name="separator">Separator between elements.</param>
        /// <param name="startWithSeparator">True to always start with a separator.</param>
        /// <param name="finishWithSeparator">True to always finish with a separator.</param>
        public static string ToListString<T>(this IEnumerable<T> list, string separator = ", ", bool startWithSeparator = false, bool finishWithSeparator = false)
        {
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteList(list, separator, startWithSeparator, finishWithSeparator);
                return stringWriter.ToString();
            }
        }

        /// <summary>
        /// Quote a content before it is written to the <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="writeContent">How to write.</param>
        /// <param name="startQuote">Left quote.</param>
        /// <param name="endQuote">Right quote.</param>
        public static void Quote(this TextWriter textWriter, Action writeContent, string startQuote = "\"", string endQuote = "\"")
        {
            textWriter.Write(startQuote);
            writeContent();
            textWriter.Write(endQuote);
        }

        /// <summary>
        /// Quote a content before it is written to the <paramref name="textWriter"/>.
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="content">What to quote.</param>
        /// <param name="startQuote">Left quote.</param>
        /// <param name="endQuote">Right quote.</param>
        public static void Quote(this TextWriter textWriter, string content, string startQuote = "\"", string endQuote = "\"")
        {
            textWriter.Write(startQuote);
            textWriter.Write(content);
            textWriter.Write(endQuote);
        }
    }
}