#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using Ami.Objects.Core;

namespace Ami.Objects
{
    /// <summary>
    /// <see cref="IObjectModel"/> selects a <see cref="IClassFactory"/> and a <see cref="IObjectFactory"/>.
    /// </summary>
    public interface IObjectModel : IDisposable
    {
        /// <summary>
        /// <see cref="IClassFactory"/> associated with this <see cref="IObjectModel"/>.
        /// </summary>
        IClassFactory ClassFactory { get; }

        /// <summary>
        /// <see cref="IObjectFactory"/> associated with this <see cref="IObjectModel"/>.
        /// </summary>
        IObjectFactory ObjectFactory { get; }

        /// <summary>
        /// <see cref="IReflector"/> associated with this <see cref="IObjectModel"/>.
        /// </summary>
        IReflector Reflector { get; }
    }

    /// <summary>
    /// <see cref="IObjectModel"/> extensions.
    /// </summary>
    public static class ObjectModelHelper
    {
        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type of domain to create, must be an interface.</typeparam>
        /// <param name="objectModel">Object model giving the <see cref="IObjectFactory"/>.</param>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        public static T Create<T>(this IObjectModel objectModel) where T : IDomain
        {
            return objectModel.ObjectFactory.CreateDomain<T>();
        }

        /// <summary>
        /// Create a new instance of <typeparamref name="T"/>.
        /// </summary>
        /// <param name="objectModel">Object model giving the <see cref="IObjectFactory"/>.</param>
        /// <param name="domain"><see cref="IObject"/>'s domain.</param>
        /// <typeparam name="T">Type of object to create, must be an interface.</typeparam>
        /// <returns>A new instance of <typeparamref name="T"/>.</returns>
        public static T Create<T>(this IObjectModel objectModel, IDomain domain) where T : IObject
        {
            return objectModel.ObjectFactory.CreateObject<T>(domain);
        }
    }
}