#region License

// ------------------------------------------------------------------------------------------- //
// The MIT License (MIT)                                                                       //
// Copyright (c) 2012 adaptive modeling interface                                              //
//                                                                                             //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this        //
// software and associated documentation files (the "Software"), to deal in the Software       //
// without restriction, including without limitation the rights to use, copy, modify, merge,   //
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons  //
// to whom the Software is furnished to do so, subject to the following conditions:            //
//                                                                                             //
// The above copyright notice and this permission notice shall be included in all copies or    //
// substantial portions of the Software.                                                       //
//                                                                                             //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,         //
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR    //
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE   //
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR        //
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER      //
// DEALINGS IN THE SOFTWARE.                                                                   //
// ------------------------------------------------------------------------------------------- //

#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

namespace Ami.Objects
{
    public static class ReflectionHelper
    {
        /// <summary>
        /// Write the property signature (PropertyType ReflectedType.PropertyName).
        /// </summary>
        /// <param name="textWriter">Where to write.</param>
        /// <param name="property">Property to write.</param>
        public static void WriteCSharpSignature(this TextWriter textWriter, PropertyInfo property)
        {
            textWriter.WriteCSharpQualifiedName(property.PropertyType);
            textWriter.Write(' ');
            textWriter.WriteCSharpQualifiedName(property.ReflectedType);
            textWriter.Write('.');
            textWriter.Write(property.Name);
        }

        /// <summary>
        /// Get the property signature (PropertyType ReflectedType.PropertyName).
        /// </summary>
        /// <param name="property">Property to write.</param>
        /// <returns>The property C# signature.</returns>
        public static string CSharpSignature(this PropertyInfo property)
        {
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteCSharpSignature(property);
                return stringWriter.ToString();
            }
        }

        /// <summary>
        /// Write the type qualified name (namespace.name).
        /// </summary>
        /// <param name="type">Type to get the qualified name from.</param>
        /// <param name="textWriter">String builder to write the name.</param>
        public static void WriteCSharpQualifiedName(this TextWriter textWriter, Type type)
        {
            if (type == null)
            {
                textWriter.Write("<null-type>");
                return;
            }
            if (type.IsGenericParameter)
            {
                textWriter.WriteCSharpTypeName(type);
                return;
            }
            if (type.IsNested)
            {
                textWriter.WriteCSharpQualifiedName(type.DeclaringType);
                textWriter.Write('.');
                textWriter.WriteCSharpName(type);
                return;
            }
            textWriter.Write(type.Namespace);
            textWriter.Write('.');
            textWriter.WriteCSharpName(type);
        }

        /// <summary>
        /// Write the type qualified name (namespace.name).
        /// </summary>
        /// <param name="type">Type to get the qualified name from.</param>
        /// <param name="textWriter">String builder to write the name.</param>
        public static void WriteCSharpName(this TextWriter textWriter, Type type)
        {
            if (type == null)
            {
                textWriter.Write("<null-type>");
                return;
            }
            if (type.IsGenericParameter)
            {
                textWriter.WriteCSharpTypeName(type);
                return;
            }
            textWriter.WriteCSharpTypeName(type);
            if (type.IsArray)
            {
                textWriter.Write('[');
                var arrayRank = type.GetArrayRank();
                for (var i = 1; i < arrayRank; ++i) textWriter.Write(',');
                textWriter.Write(']');
                return;
            }
        }

        private static void WriteCSharpTypeName(this TextWriter textWriter, Type type)
        {
            if (type.IsGenericParameter)
            {
                textWriter.Write(type.Name);
                return;
            }
            var typeName = type.Name;
            if (!type.IsGenericType)
            {
                if (!type.IsArray)
                {
                    textWriter.Write(typeName);
                    return;
                }
                textWriter.Write(typeName.Substring(0, typeName.Length - 2));
                return;
            }
            textWriter.Write(typeName.Substring(0, typeName.IndexOf('`')));
            textWriter.Write('<');
            var genericArguements = type.GetGenericArguments();
            var count = 0;
            foreach (var genericArgument in genericArguements)
            {
                if (count != 0) textWriter.Write(", ");
                textWriter.WriteCSharpQualifiedName(genericArgument);
                ++count;
            }
            textWriter.Write('>');
        }


        /// <summary>
        /// Create the type qualified name (namespace.name).
        /// </summary>
        /// <param name="type">Type to get the qualified name from.</param>
        /// <returns>Type qualified name as in C# conventions.</returns>
        public static string CSharpQualifiedName(this Type type)
        {
            using (var stringWriter = new StringWriter())
            {
                stringWriter.WriteCSharpQualifiedName(type);
                return stringWriter.ToString();
            }
        }

        /// <summary>
        /// Changes the name of the given <paramref name="type"/> for <paramref name="newName"/>.
        /// </summary>
        /// <param name="type">Original type.</param>
        /// <param name="newName">New type name.</param>
        /// <returns>The C# qualified name.</returns>
        public static string ChangeCSharpQualifiedName(this Type type, string newName)
        {
            var stringWriter = new StringWriter();
            stringWriter.Write(type.Namespace);
            var nestedType = type.IsNested ? type.DeclaringType : null;
            while (nestedType != null)
            {
                stringWriter.Write('.');
                stringWriter.WriteCSharpName(nestedType);
                nestedType = nestedType.IsNested ? nestedType.DeclaringType : null;
            }
            stringWriter.Write('.');
            stringWriter.Write(newName);
            return stringWriter.ToString();
        }

        /// <summary>
        /// Get the most common <typeparamref name="T"/> by comparing <paramref name="getType"/> values on each item.
        /// The item returned is the one with the most common type.
        /// </summary>
        /// <typeparam name="T">Type of value to select.</typeparam>
        /// <param name="values">Values.</param>
        /// <param name="getType">Get a type from a value.</param>
        /// <returns>The most common type.</returns>
        public static T CommonType<T>(this IEnumerable<T> values, Func<T, Type> getType) where T : class
        {
            var commons = values.Where(v => values.All(otherV => otherV == v || getType(v).Inherits(getType(otherV)))).Distinct();
            if (commons.Any())
            {
                if (commons.Skip(1).Any()) return null;
                return commons.First();
            }
            return null;
        }

        /// <summary>
        /// Checks that the given <paramref name="type"/> inherits <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Type to inherit.</typeparam>
        /// <param name="type">Typed checked.</param>
        /// <param name="strict">When set to true, a type doesn't inherit itself.</param>
        /// <returns>True if <paramref name="type"/> inherist <typeparamref name="T"/>, otherwise false.</returns>
        public static bool Inherits<T>(this Type type, bool strict = false)
        {
            return Inherits(type, typeof (T), strict);
        }


        /// <summary>
        /// Checks that the given <paramref name="type"/> inherits <param name="typeToInherit"></param>.
        /// </summary>
        /// <param name="type">Typed checked.</param>
        /// <param name="typeToInherit">Type to inherit.</param>
        /// <param name="strict">When set to true, a type doesn't inherit itself.</param>
        /// <returns>True if <paramref name="type"/> inherist <param name="typeToInherit"></param>, otherwise false.</returns>
        public static bool Inherits(this Type type, Type typeToInherit, bool strict = false)
        {
            if (strict && ReferenceEquals(typeToInherit, type)) return false;
            return typeToInherit.IsAssignableFrom(type);
        }

        /// <summary>
        /// Get all direct interfaces of the given <paramref name="type"/>.
        /// </summary>
        /// <param name="type">Type.</param>
        /// <returns>All direct interfaces of the given <paramref name="type"/>.</returns>
        public static IEnumerable<Type> DirectInterfaces(this Type type)
        {
            var allInterfaces = type.GetInterfaces();
            return allInterfaces.Where(i => !allInterfaces.Where(otherI => !ReferenceEquals(otherI, i)).Any(otherI => otherI.Inherits(i)));
        }

        /// <summary>
        /// Get the first attribute of a given type.
        /// </summary>
        /// <typeparam name="T">Attribute type.</typeparam>
        /// <param name="o">Object to get the attributes from.</param>
        /// <param name="strict">True to get only instance of the given type, not sub types.</param>
        /// <returns>The first attribute of the given type or null if not found.</returns>
        public static T GetFirstAttribute<T>(this ICustomAttributeProvider o, bool strict = false) where T : Attribute
        {
            return (T) o.GetFirstAttribute(typeof (T), strict);
        }

        /// <summary>
        /// Get the first attribute of a given type.
        /// </summary>
        /// <param name="o">Object to get the attributes from.</param>
        /// <param name="attributeType">Attribute type.</param>
        /// <param name="strict">True to get only instance of the given type, not sub types.</param>
        /// <returns>The first attribute of the given type or null if not found.</returns>
        public static Attribute GetFirstAttribute(this ICustomAttributeProvider o, Type attributeType, bool strict = false)
        {
            var attributes = o.GetCustomAttributes(attributeType, false);
            if (strict) return (Attribute) attributes.FirstOrDefault(a => a.GetType() == attributeType);
            if (attributes.Length > 0) return (Attribute) attributes[0];
            return null;
        }

        /// <summary>
        /// Check if an object has a given <typeparamref name="T"/>.
        /// </summary>
        /// <typeparam name="T">Attribute type.</typeparam>
        /// <param name="o">Object to check.</param>
        /// <returns>True if the object has at least one attribute of type <typeparamref name="T"/>.</returns>
        public static bool HasAttribute<T>(this ICustomAttributeProvider o) where T : Attribute
        {
            return o.HasAttribute(typeof (T));
        }

        /// <summary>
        /// Check if an object has a given <paramref name="attributeType"/>.
        /// </summary>
        /// <param name="o">Object to check.</param>
        /// <param name="attributeType">Attribute type.</param>
        /// <returns>True if the object has at least one attribute of type <paramref name="attributeType"/>.</returns>
        public static bool HasAttribute(this ICustomAttributeProvider o, Type attributeType)
        {
            return o.GetCustomAttributes(attributeType, false).Length > 0;
        }

        /// <summary>
        /// Check if a <paramref name="type"/> has a property called <paramref name="propertyName"/>.
        /// </summary>
        /// <param name="type">Type to check.</param>
        /// <param name="propertyName">Name of the property to check.</param>
        /// <returns>True if the <paramref name="type"/> has a property called <paramref name="propertyName"/>.</returns>
        public static bool HasProperty(this Type type, string propertyName)
        {
            return type.GetProperty(propertyName) != null;
        }

        /// <summary>
        /// Find a type by its C# qualifed name.
        /// </summary>
        /// <param name="assembly">Assembly where to look.</param>
        /// <param name="csharpQualifiedName">C# qualifed name of the type to find.</param>
        /// <returns>The type or null if not found.</returns>
        public static Type FindTypeByCSharpQualifiedName(this Assembly assembly, string csharpQualifiedName)
        {
            var stringBuilder = new StringBuilder();
            using (var stringWriter = new StringWriter(stringBuilder))
            {
                return assembly.GetTypes().FirstOrDefault(t =>
                                                              {
                                                                  stringBuilder.Clear();
                                                                  stringWriter.WriteCSharpQualifiedName(t);
                                                                  return stringWriter.ToString() == csharpQualifiedName;
                                                              });
            }
        }
    }
}